from __future__ import annotations

import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, Iterator, Mapping, cast, overload

from ..core import AbstractYAMLObjectSingleton, Frame, storage_err_msg
from . import utils
from .abstract import AbstractPlugin, PluginConfig
from .utils import *  # noqa: F401,F403

if TYPE_CHECKING:
    from ..main import MainWindow


__all__ = [
    'AbstractPlugin', 'PluginConfig',
    *utils.__all__
]


class Plugins(AbstractYAMLObjectSingleton):
    __slots__ = ()

    _closure = {**globals()}

    # tab idx, clip idx, frame
    last_frame_change: tuple[int, int, int]

    # tab idx, clip idx
    last_output_change: tuple[int, int]

    @classmethod
    def file_to_plugins(cls, path: Path) -> Iterable[type[AbstractPlugin]]:
        from importlib.util import module_from_spec, spec_from_file_location
        if True:
            # has to be imported after the main
            from importlib._bootstrap_external import SOURCE_SUFFIXES  # type: ignore
            SOURCE_SUFFIXES.append('.ppy')

        spec = spec_from_file_location(path.stem, path)
        module = module_from_spec(spec)

        spec.loader.exec_module(module)

        try:
            module_all = object.__getattribute__(module, '__all__')
        except AttributeError:
            print(ImportWarning(f'The plugin "{path.stem}" has no __all__ defined and thus can\'t be imported!'))
            return

        for export in cast(tuple[str, ...], module_all):
            exp_obj = object.__getattribute__(module, export)

            if not isinstance(exp_obj, type):
                continue

            if not issubclass(exp_obj, AbstractPlugin):
                continue

            if not hasattr(exp_obj, '_config'):
                print(ImportWarning(
                    f'The plugin "{exp_obj.__name__}" has no config set up and thus can\'t be imported!'
                ))
                continue

            yield exp_obj

    def __init__(self, main: MainWindow) -> None:
        main.plugins = self

        self.main = main
        self.plugins_tab = main.plugins_tab
        self.main.main_split.setSizes([0, 0])
        self.main.reload_before_signal.connect(self.reset_last_reload)

        self.reset_last_reload()

        self.plugins = dict[str, AbstractPlugin]()

        plugin_files = list[Path]()

        def _find_files(folder: Path, ignore_path: bool = False) -> None:
            files = list(folder.glob('*.ppy'))

            if files:
                if not ignore_path:
                    sys.path.append(str(folder))

                plugin_files.extend(files)

        def _check_folder(folder: str | Path, ignore_path: bool = False) -> None:
            if not folder:
                return

            if isinstance(folder, str):
                folder = folder.strip()

                if folder.startswith(';'):
                    return

                folder = Path(folder)

            if not folder.is_dir():
                return

            _find_files(folder, ignore_path)

            for folder in (f for f in folder.glob('*') if f.is_dir()):
                _find_files(folder, ignore_path)

        _check_folder(Path(__file__).parent, True)
        _check_folder(self.main.global_plugins_dir)

        for paths_file in self.main.global_plugins_dir.glob('*.pth'):
            for path in paths_file.read_text('utf8').splitlines():
                _check_folder(path)

        for plugin_file in plugin_files:
            for plugin in self.file_to_plugins(plugin_file):
                if plugin._config.namespace in self.plugins:
                    print(UserWarning(
                        f'Tried to register plugin "{plugin._config.display_name} '
                        f'({plugin._config.namespace})" twice!'
                    ))

                self.plugins[plugin._config.namespace] = plugin(self.main)

        i = 0
        for name, plugin in self.plugins.items():
            plugin.setObjectName(f'Plugins.{name}')

            if not plugin._config.visible_in_tab:
                try:
                    plugin.hide()
                except Exception:
                    ...

                continue

            plugin.index = i

            self.plugins_tab.addTab(plugin, plugin._config.display_name)
            i += 1

    def reset_last_reload(self) -> None:
        self.last_frame_change = (-1, -1, -1)

        self.last_output_change = (-1, -1)

    def init_outputs(self) -> None:
        for plugin in self:
            plugin.init_outputs()

    def on_current_frame_changed(self, frame: Frame) -> None:
        tab_idx = self.plugins_tab.currentIndex()

        curr_render = (tab_idx, self.main.current_output.index, int(frame))

        if self.main.main_split.current_position and self.last_frame_change != curr_render:
            self[tab_idx].on_current_frame_changed(frame)

            self.last_frame_change = curr_render

    def on_current_output_changed(self, index: int, prev_index: int) -> None:
        tab_idx = self.plugins_tab.currentIndex()

        curr_output = (tab_idx, index)

        if self.main.main_split.current_position and self.last_output_change != curr_output:
            self[tab_idx].on_current_output_changed(index, prev_index)

            self.last_output_change = curr_output

    def update(self, frame: Frame | None = None, index: int | None = None, prev_index: int | None = None) -> None:
        from vstools import fallback

        if not self.main.current_output:
            return

        self.on_current_frame_changed(
            fallback(frame, self.main.current_output.last_showed_frame)
        )
        self.on_current_output_changed(
            fallback(index, self.main.current_output.index),
            fallback(prev_index, self.main.current_output.index),
        )

        sizel, sizer = self.main.main_split.sizes()

        self.main.main_split.setSizes([sizel, sizer - 1])
        self.main.main_split.setSizes([sizel, sizer])

    @overload
    def __getitem__(self, _sub: str | int) -> AbstractPlugin:
        ...

    @overload
    def __getitem__(self, _sub: slice) -> list[AbstractPlugin]:
        ...

    def __getitem__(self, _sub: str | int | slice) -> AbstractPlugin | list[AbstractPlugin]:
        length = len(self.plugins)
        if isinstance(_sub, slice):
            return [self[i] for i in range(*_sub.indices(length))]

        if isinstance(_sub, int):
            if _sub < 0:
                _sub += length
            if _sub < 0 or _sub >= length:
                raise IndexError

            _sub = list(self.plugins.keys())[_sub]

        return cast(AbstractPlugin, self.plugins[_sub])

    def __len__(self) -> int:
        return len(self.plugins)

    if TYPE_CHECKING:
        # https://github.com/python/mypy/issues/2220
        def __iter__(self) -> Iterator[AbstractPlugin]:
            ...

    def __getstate__(self) -> Mapping[str, Mapping[str, Any]]:
        return {
            toolbar_name: getattr(self, toolbar_name).__getstate__()
            for toolbar_name in self.plugins
        }

    def __setstate__(self, state: Mapping[str, Mapping[str, Any]]) -> None:
        for toolbar_name in self.plugins:
            try:
                storage = state[toolbar_name]
                if not isinstance(storage, Mapping):
                    raise TypeError
                getattr(self, toolbar_name).__setstate__(storage)
            except (KeyError, TypeError) as e:
                logging.error(e)
                logging.warning(storage_err_msg(toolbar_name))



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, overload

from PyQt6.QtWidgets import QSizePolicy, QWidget
from vstools import vs, vs_object

from ..core import Frame, GraphicsView, VideoOutput, VideoOutputNode, main_window
from .abstract import AbstractPlugin

if TYPE_CHECKING:
    from ..main import MainWindow

__all__ = [
    'PluginVideoOutputs', 'LazyMapPluginVideoOutputs',

    'MappedNodesPlugin', 'MappedNodesViewPlugin',

    'PluginGraphicsView'
]


class PluginVideoOutputs(vs_object, dict[VideoOutput, VideoOutput]):
    def __init__(self, main: MainWindow | None = None) -> None:
        self.main = main or main_window()

    @property
    def current(self) -> VideoOutput:
        return self[self.main.current_output]

    @property
    def source(self) -> vs.VideoNode:
        return self.current.source.clip

    @property
    def prepared(self) -> vs.VideoNode:
        return self.current.prepared.clip

    def __delitem__(self, __key: VideoOutput) -> None:
        if __key not in self:
            return

        return super().__delitem__(__key)

    def __vs_del__(self, core_id: int) -> None:
        self.clear()


class LazyMapPluginVideoOutputs(PluginVideoOutputs):
    def __init__(self, main: MainWindow | None, func: Callable[[vs.VideoNode], vs.VideoNode]) -> None:
        super().__init__(main)
        self.func = func

    def __getitem__(self, key: VideoOutput) -> VideoOutput:
        if key not in self:
            self[key] = key.with_node(self.func(key.source.clip))

        return super().__getitem__(key)


class MappedNodesPlugin(AbstractPlugin):
    def __init__(self, main: MainWindow) -> None:
        super().__init__(main)

        self.outputs = LazyMapPluginVideoOutputs(main, self.get_node)

    @overload
    def get_node(self, node: vs.VideoNode) -> vs.VideoNode:
        ...

    @overload
    def get_node(self, node: vs.VideoNode) -> VideoOutputNode:
        ...

    @overload
    def get_node(self, node: vs.VideoNode) -> vs.VideoNode | VideoOutputNode:
        ...

    def get_node(self, node: vs.VideoNode) -> vs.VideoNode | VideoOutputNode:
        raise NotImplementedError

    def init_outputs(self) -> None:
        assert self.main.outputs

        self.outputs.clear()

    def reset(self) -> None:
        self.init_outputs()

        self.on_current_frame_changed(self.main.current_output.last_showed_frame)

    def on_current_frame_changed(self, frame: Frame) -> None:
        raise NotImplementedError


class PluginGraphicsView(GraphicsView):
    plugin: MappedNodesPlugin

    def __init__(
        self, main: MainWindow | MappedNodesPlugin, plugin: MappedNodesPlugin | None = None,
        parent: QWidget | None = None
    ) -> None:
        if isinstance(main, MappedNodesPlugin):
            plugin = main
            main = main.main

        super().__init__(main, parent)

        if plugin:
            self.plugin = plugin
            self.setSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)
        elif isinstance(self, MappedNodesPlugin):
            self.plugin = self
        else:
            raise RuntimeError('This GraphicsView has to be bound to a MappedNodesPlugin!')

    @property
    def content_width(self) -> int:
        return self.plugin.outputs.current.width

    @property
    def content_height(self) -> int:
        return self.plugin.outputs.current.height


class MappedNodesViewPlugin(MappedNodesPlugin, PluginGraphicsView):
    def on_current_frame_changed(self, frame: Frame) -> None:
        self.outputs.current.render_frame(frame, None, None, self.current_scene)



from __future__ import annotations

from copy import deepcopy
from typing import Any, cast

from PyQt6.QtCore import QEvent, QLineF, QRectF, Qt, pyqtSignal
from PyQt6.QtGui import QMouseEvent, QMoveEvent, QPainter, QPaintEvent, QPalette, QPen, QResizeEvent
from PyQt6.QtWidgets import QApplication, QToolTip, QWidget

from ..core import NotchProvider, AbstractYAMLObject, Frame, Notch, Notches, Time, VideoOutput, main_window
from ..utils import strfdelta

__all__ = [
    'Timeline'
]


class Timeline(QWidget):
    __slots__ = (
        'app', 'main',
        'rectF', 'prevRectF',
        'totalT', 'totalF',
        'notchIntervalTargetX', 'notchHeight', 'fontHeight',
        'notchLabelInterval', 'notchScrollInterval', 'scrollHeight',
        'cursorX', 'cursorFT', 'needFullRepaint',
        'scrollRect',
    )

    class Mode(AbstractYAMLObject):
        FRAME = 'frame'
        TIME = 'time'

        @classmethod
        def is_valid(cls, value: str) -> bool:
            return value in {cls.FRAME, cls.TIME}

    clicked = pyqtSignal(Frame, Time)

    def __init__(self, parent: QWidget, **kwargs: Any) -> None:
        super().__init__(parent, **kwargs)
        self.app = QApplication.instance()
        self.main = main_window()

        self._mode = self.Mode.TIME

        self.rect_f = QRectF()

        self.end_t = Time(seconds=1)
        self.end_f = Frame(1)

        self.notch_interval_target_x = round(75 * self.main.display_scale)
        self.notch_height = round(6 * self.main.display_scale)
        self.font_height = round(10 * self.main.display_scale)
        self.notch_label_interval = round(-1 * self.main.display_scale)
        self.notch_scroll_interval = round(2 * self.main.display_scale)
        self.scroll_height = round(10 * self.main.display_scale)

        self.setMinimumSize(self.notch_interval_target_x, round(33 * self.main.display_scale))

        font = self.font()
        font.setPixelSize(self.font_height)
        self.setFont(font)

        self.cursor_x = 0
        # used as a fallback when self.rectF.width() is 0,
        # so cursorX is incorrect
        self.cursor_ftx: Frame | Time | int | None = None
        # False means that only cursor position'll be recalculated
        self.need_full_repaint = True

        self.notches = dict[NotchProvider, Notches]()

        self.setAttribute(Qt.WidgetAttribute.WA_OpaquePaintEvent)
        self.setMouseTracking(True)

    def paintEvent(self, event: QPaintEvent) -> None:
        super().paintEvent(event)
        self.rect_f = QRectF(event.rect())
        # self.rectF.adjust(0, 0, -1, -1)

        if self.cursor_ftx is not None:
            self.set_position(self.cursor_ftx)
        self.cursor_ftx = None

        painter = QPainter(self)
        self.drawWidget(painter)

    def drawWidget(self, painter: QPainter) -> None:
        if self.need_full_repaint:
            labels_notches = Notches()
            label_notch_bottom = (
                self.rect_f.top() + self.font_height + self.notch_label_interval + self.notch_height + 5
            )
            label_notch_top = label_notch_bottom - self.notch_height
            label_notch_x = self.rect_f.left()

            if self.mode == self.Mode.TIME:
                notch_interval_t = self.calculate_notch_interval_t(self.notch_interval_target_x)
                label_format = self.generate_label_format(notch_interval_t, self.end_t)
                label_notch_t = Time()

                while (label_notch_x < self.rect_f.right() and label_notch_t <= self.end_t):
                    line = QLineF(label_notch_x, label_notch_bottom, label_notch_x, label_notch_top)
                    labels_notches.add(Notch(deepcopy(label_notch_t), line=line))
                    label_notch_t += notch_interval_t
                    label_notch_x = self.t_to_x(label_notch_t)

            elif self.mode == self.Mode.FRAME:
                notch_interval_f = self.calculate_notch_interval_f(self.notch_interval_target_x)
                label_notch_f = Frame(0)

                while (label_notch_x < self.rect_f.right() and label_notch_f <= self.end_f):
                    line = QLineF(label_notch_x, label_notch_bottom, label_notch_x, label_notch_top)
                    labels_notches.add(Notch(deepcopy(label_notch_f), line=line))
                    label_notch_f += notch_interval_f
                    label_notch_x = self.f_to_x(label_notch_f)

            self.scroll_rect = QRectF(
                self.rect_f.left(),
                label_notch_bottom + self.notch_scroll_interval,
                self.rect_f.width(), self.scroll_height
            )

            for provider, notches in self.notches.items():
                if not provider.is_notches_visible:
                    continue

                for notch in notches:
                    if isinstance(notch.data, Frame):
                        x = self.f_to_x(notch.data)
                    elif isinstance(notch.data, Time):
                        x = self.t_to_x(notch.data)
                    y = self.scroll_rect.top()
                    notch.line = QLineF(x, y, x, y + self.scroll_rect.height() - 1)

        cursor_line = QLineF(
            self.cursor_x, self.scroll_rect.top(), self.cursor_x,
            self.scroll_rect.top() + self.scroll_rect.height() - 1
        )

        # drawing

        if self.need_full_repaint:
            painter.fillRect(self.rect_f, self.palette().color(QPalette.ColorRole.Window))

            painter.setPen(QPen(self.palette().color(QPalette.ColorRole.WindowText)))
            painter.setRenderHint(QPainter.RenderHint.Antialiasing, False)
            painter.drawLines([notch.line for notch in labels_notches])  # type: ignore
            painter.setRenderHint(QPainter.RenderHint.Antialiasing)

            for i, notch in enumerate(labels_notches):
                line = notch.line
                anchor_rect = QRectF(
                    line.x2(), line.y2() - self.notch_label_interval, 0, 0)

                if self.mode == self.Mode.TIME:
                    time = cast(Time, notch.data)
                    label = strfdelta(time, label_format)
                if self.mode == self.Mode.FRAME:
                    label = str(notch.data)

                if i == 0:
                    rect = painter.boundingRect(
                        anchor_rect, Qt.AlignmentFlag.AlignBottom | Qt.AlignmentFlag.AlignLeft, label
                    )
                    if self.mode == self.Mode.TIME:
                        rect.moveLeft(-2.5)
                elif i == (len(labels_notches) - 1):
                    rect = painter.boundingRect(
                        anchor_rect, Qt.AlignmentFlag.AlignBottom | Qt.AlignmentFlag.AlignHCenter, label
                    )
                    if rect.right() > self.rect_f.right():
                        rect = painter.boundingRect(
                            anchor_rect, Qt.AlignmentFlag.AlignBottom | Qt.AlignmentFlag.AlignRight, label
                        )
                else:
                    rect = painter.boundingRect(
                        anchor_rect, Qt.AlignmentFlag.AlignBottom | Qt.AlignmentFlag.AlignHCenter, label
                    )
                painter.drawText(rect, label)

        painter.setRenderHint(QPainter.RenderHint.Antialiasing, False)
        painter.fillRect(self.scroll_rect, Qt.GlobalColor.gray)

        for provider, notches in self.notches.items():
            if not provider.is_notches_visible:
                continue

            for notch in notches:
                painter.setPen(notch.color)
                painter.drawLine(notch.line)

        painter.setPen(Qt.GlobalColor.black)
        painter.drawLine(cursor_line)

        self.need_full_repaint = False

    def full_repaint(self) -> None:
        self.need_full_repaint = True
        self.update()

    def moveEvent(self, event: QMoveEvent) -> None:
        super().moveEvent(event)
        self.full_repaint()

    def mousePressEvent(self, event: QMouseEvent) -> None:
        super().mousePressEvent(event)
        pos = event.pos().toPointF()
        if self.scroll_rect.contains(pos):
            self.set_position(int(pos.x()))
            self.clicked.emit(self.x_to_f(self.cursor_x, Frame), self.x_to_t(self.cursor_x, Time))

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        super().mouseMoveEvent(event)
        for provider, notches in self.notches.items():
            if not provider.is_notches_visible:
                continue
            for notch in notches:
                line = notch.line
                if line.x1() - 0.5 <= event.pos().x() <= line.x1() + 0.5:
                    QToolTip.showText(event.globalPosition().toPoint(), notch.label)
                    return

    def resizeEvent(self, event: QResizeEvent) -> None:
        super().resizeEvent(event)
        self.full_repaint()

    def event(self, event: QEvent) -> bool:
        if event.type() in {QEvent.Type.Polish, QEvent.Type.ApplicationPaletteChange}:
            self.setPalette(self.main.palette())
            self.full_repaint()
            return True

        return super().event(event)

    def update_notches(self, provider: NotchProvider | None = None) -> None:
        if provider is not None:
            self.notches[provider] = provider.get_notches()
        else:
            for t in self.main.toolbars:
                self.notches[t] = t.get_notches()
            for t in self.main.plugins:
                self.notches[t] = t.get_notches()
        self.full_repaint()

    @property
    def mode(self) -> str:
        return self._mode

    @mode.setter
    def mode(self, value: str) -> None:
        if value == self._mode:
            return

        self._mode = value
        self.full_repaint()

    notch_intervals_t = list(
        Time(seconds=n) for n in [
            1, 2, 5, 10, 15, 30, 60, 90, 120, 300, 600,
            900, 1200, 1800, 2700, 3600, 5400, 7200
        ]
    )

    def calculate_notch_interval_t(self, target_interval_x: int) -> Time:
        margin = 1 + self.main.settings.timeline_label_notches_margin / 100
        target_interval_t = self.x_to_t(target_interval_x, Time)
        if target_interval_t >= self.notch_intervals_t[-1] * margin:
            return self.notch_intervals_t[-1]
        for interval in self.notch_intervals_t:
            if target_interval_t < interval * margin:
                return interval
        raise RuntimeError

    notch_intervals_f = list(
        Frame(n) for n in [
            1, 5, 10, 20, 25, 50, 75, 100, 200, 250, 500, 750, 1000,
            2000, 2500, 5000, 7500, 10000, 20000, 25000, 50000, 75000
        ]
    )

    def calculate_notch_interval_f(self, target_interval_x: int) -> Frame:
        margin = 1 + self.main.settings.timeline_label_notches_margin / 100
        target_interval_f = self.x_to_f(target_interval_x, Frame)
        if target_interval_f >= Frame(
                round(int(self.notch_intervals_f[-1]) * margin)):
            return self.notch_intervals_f[-1]
        for interval in self.notch_intervals_f:
            if target_interval_f < Frame(
                    round(int(interval) * margin)):
                return interval
        raise RuntimeError

    def generate_label_format(self, notch_interval_t: Time, end_time: Time | Time) -> str:
        if end_time >= Time(hours=1):
            return '%h:%M:00'
        elif notch_interval_t >= Time(minutes=1):
            return '%m:00'
        else:
            return '%m:%S'

    def set_end_frame(self, node: VideoOutput) -> None:
        self.end_f = node.total_frames
        self.end_t = node.total_time
        self.full_repaint()

    def set_position(self, pos: Frame | Time | int) -> None:
        if self.rect_f.width() == 0.0:
            self.cursor_ftx = pos

        if isinstance(pos, Frame):
            self.cursor_x = self.f_to_x(pos)
        elif isinstance(pos, Time):
            self.cursor_x = self.t_to_x(pos)
        elif isinstance(pos, int):
            self.cursor_x = pos
        else:
            raise TypeError
        self.update()

    def t_to_x(self, t: Time) -> int:
        width = self.rect_f.width()
        try:
            x = round(float(t) / float(self.end_t) * width)
        except ZeroDivisionError:
            x = 0
        return x

    def x_to_t(self, x: int, ty: type[Time]) -> Time:
        width = self.rect_f.width()
        return ty(seconds=(x * float(self.end_t) / width))

    def f_to_x(self, f: Frame) -> int:
        width = self.rect_f.width()
        try:
            x = round(int(f) / int(self.end_f) * width)
        except ZeroDivisionError:
            x = 0
        return x

    def x_to_f(self, x: int, ty: type[Frame]) -> Frame:
        width = self.rect_f.width()
        value = round(x / width * int(self.end_f))
        return ty(value)



from __future__ import annotations

from .info import is_preview

__all__ = [
    'start_preview'
]


def start_preview(path: str, *args: str) -> None:
    from ..init import main

    if is_preview():
        from vstools import CustomRuntimeError

        raise CustomRuntimeError('You can\'t use this function from a file getting previewed!')

    main([path, *args], True)



from __future__ import annotations

import io
import logging
import sys
from fractions import Fraction
from functools import partial
from importlib import reload as reload_module
from pathlib import Path
from time import time
from typing import Any, Iterable, Mapping, cast

import vapoursynth as vs
from PyQt6 import QtCore
from PyQt6.QtCore import QEvent, QKeyCombination, QRectF, Qt, pyqtSignal
from PyQt6.QtGui import QCloseEvent, QColorSpace, QKeySequence, QMoveEvent, QShortcut, QShowEvent
from PyQt6.QtWidgets import QApplication, QLabel, QMainWindow, QSizePolicy, QSplitter, QTabWidget
from vsengine import vpy  # type: ignore

from ..core import (
    PRELOADED_MODULES, AbstractQItem, CroppingInfo, DragNavigator, ExtendedWidget, Frame, GraphicsImageItem,
    GraphicsView, HBoxLayout, MainVideoOutputGraphicsView, QAbstractYAMLObjectSingleton, StatusBar, Time, Timer,
    VBoxLayout, VideoOutput, _monkey_runpy_dicts, dispose_environment, get_current_environment, make_environment
)
from ..models import GeneralModel, VideoOutputs
from ..plugins import Plugins
from ..toolbars import Toolbars
from ..utils import fire_and_forget, set_status_label
from .dialog import ScriptErrorDialog, SettingsDialog
from .settings import MainSettings, WindowSettings
from .timeline import Timeline

if sys.platform == 'win32':
    try:
        import win32gui  # type: ignore[import]
        from PIL import _imagingcms  # type: ignore[attr-defined]
    except ImportError:
        _imagingcms = None

    from os.path import expandvars
else:
    from os.path import expanduser

try:
    from yaml import CDumper as yaml_Dumper
    from yaml import CLoader as yaml_Loader
except ImportError:
    from yaml import Dumper as yaml_Dumper  # type: ignore
    from yaml import Loader as yaml_Loader  # type: ignore

from yaml import MarkedYAMLError, YAMLError

__all__ = [
    'MainWindow'
]


class CentralSplitter(QSplitter):
    def __init__(self, main_window: MainWindow, orientation: QtCore.Qt.Orientation) -> None:
        super().__init__(orientation)

        self.main_window = main_window

        self.splitterMoved.connect(self.on_splitter_moved)

        self.previous_position = 0

    @property
    def current_position(self) -> int:
        return self.sizes()[-1]

    def on_splitter_moved(self) -> None:
        if self.previous_position == 0 and self.current_position:
            self.main_window.plugins.update()

        self.previous_position = self.current_position


class MainWindow(AbstractQItem, QMainWindow, QAbstractYAMLObjectSingleton):
    VSP_DIR_NAME = '.vspreview'
    VSP_GLOBAL_DIR_NAME = Path(
        expandvars('%APPDATA%') if sys.platform == "win32" else expanduser('~/.config')  # type: ignore
    )

    global_config_dir = VSP_GLOBAL_DIR_NAME / VSP_DIR_NAME
    global_storage_path = global_config_dir / '.global.yml'
    global_plugins_dir = global_config_dir / 'plugins'

    VSP_VERSION = 3.2
    BREAKING_CHANGES_VERSIONS = list[str](['3.0', '3.1'])

    # status bar
    def STATUS_FRAME_PROP(self, prop: Any) -> str:
        return 'Type: %s' % (prop['_PictType'].decode('utf-8') if '_PictType' in prop else '?')

    EVENT_POLICY = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)

    storable_attrs = ('settings', 'toolbars')

    __slots__ = (
        *storable_attrs, 'app', 'display_scale', 'clipboard',
        'script_path', 'timeline', 'main_layout', 'autosave_timer',
        'graphics_view', 'script_error_dialog',
        'central_widget', 'statusbar', 'storage_not_found',
        'current_storage_path'
    )

    # emit when about to reload a script: clear all existing references to existing clips.
    reload_signal = pyqtSignal()
    reload_before_signal = pyqtSignal()
    reload_after_signal = pyqtSignal()
    cropValuesChanged = pyqtSignal(CroppingInfo)

    toolbars: Toolbars
    plugins: Plugins
    app_settings: SettingsDialog
    window_settings = WindowSettings()

    autosave_timer: Timer

    def __init__(self, config_dir: Path, no_exit: bool) -> None:
        from ..toolbars import MainToolbar

        super().__init__()

        self.no_exit = no_exit

        self.settings = MainSettings(MainToolbar)

        self.current_config_dir = config_dir / self.VSP_DIR_NAME
        self.global_plugins_dir.mkdir(parents=True, exist_ok=True)

        self.app = cast(QApplication, QApplication.instance())
        assert self.app

        self.last_reload_time = time()

        self.bound_graphics_views = dict[GraphicsView, set[GraphicsView]]()

        if self.settings.dark_theme_enabled:
            from ..core import apply_plotting_style

            try:
                from qdarkstyle import _load_stylesheet  # type: ignore[import]
            except ImportError:
                self.settings.dark_theme_enabled = False
            else:
                apply_plotting_style()
                self.app.setStyleSheet(self.patch_dark_stylesheet(_load_stylesheet(qt_api='pyqt6')))

        self.ensurePolished()

        self.display_scale = self.app.primaryScreen().logicalDotsPerInch() / self.settings.base_ppi
        self.setWindowTitle('VSPreview')

        desktop_size = self.app.primaryScreen().size()

        self.move(int(desktop_size.width() * 0.15), int(desktop_size.height() * 0.075))
        self.setup_ui()
        self.storage_not_found = False
        self.timecodes = dict[
            int, tuple[str | Path | dict[
                tuple[int | None, int | None], float | tuple[int, int] | Fraction
            ] | list[Fraction], int | None]
        ]()
        self.norm_timecodes = dict[int, list[float]]()

        self.user_output_info = {
            vs.VideoNode: dict[int, dict[str, Any]](),
            vs.AudioNode: dict[int, dict[str, Any]](),
            vs.RawNode: dict[int, dict[str, Any]]()
        }

        # global
        self.clipboard = self.app.clipboard()
        self.external_args = list[tuple[str, str]]()
        self.script_path = Path()
        self.script_exec_failed = False
        self.current_storage_path = Path()

        # timeline
        self.timeline.clicked.connect(self.on_timeline_clicked)

        # display profile
        self.display_profile: QColorSpace | None = None
        self.current_screen = self.app.primaryScreen()

        # init toolbars and outputs
        self.app_settings = SettingsDialog(self)

        Toolbars(self)

        for toolbar in self.toolbars:
            self.main_layout.addWidget(toolbar)
            self.toolbars.main.layout().addWidget(toolbar.toggle_button)

        Plugins(self)

        self.app_settings.tab_widget.setUsesScrollButtons(False)
        self.app_settings.setMinimumWidth(
            int(len(self.toolbars) * 1.05 * self.app_settings.tab_widget.geometry().width() / 2)
        )

        self.set_qobject_names()
        self.setObjectName('MainWindow')

        self.env: vpy.Script | None = None

    def setup_ui(self) -> None:
        self.central_widget = ExtendedWidget(self)
        self.main_layout = VBoxLayout(self.central_widget)
        self.setCentralWidget(self.central_widget)

        self.graphics_view = MainVideoOutputGraphicsView(self, self.central_widget)

        DragNavigator(self, self.graphics_view)

        self.timeline = Timeline(self.central_widget)

        self.plugins_tab = QTabWidget(self.central_widget)
        self.plugins_tab.currentChanged.connect(lambda x: self.plugins.update())

        self.main_split = CentralSplitter(self, QtCore.Qt.Orientation.Horizontal)
        self.main_split.addWidget(self.graphics_view)
        self.main_split.addWidget(self.plugins_tab)
        self.main_split.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)

        HBoxLayout(self.main_layout, [self.main_split])
        self.main_layout.addWidget(self.timeline)

        # status bar
        self.statusbar = StatusBar(self.central_widget)

        for name in self.statusbar.label_names:
            self.statusbar.__setattr__(name, QLabel(self.central_widget))

        self.statusbar.addWidgets([
            getattr(self.statusbar, name) for name in self.statusbar.label_names
        ])

        self.statusbar.addPermanentWidget(self.statusbar.label)

        self.setStatusBar(self.statusbar)

        # dialogs
        self.script_error_dialog = ScriptErrorDialog(self)

        self.autosave_timer = Timer(timeout=self.dump_storage_async)
        self.reload_signal.connect(self.autosave_timer.stop)

        QShortcut(
            QKeySequence(QKeyCombination(Qt.Modifier.CTRL, Qt.Key.Key_A).toCombined()),
            self, activated=self.auto_fit_keyswitch
        )

    def auto_fit_keyswitch(self) -> None:
        for view in self.graphics_views:
            if view.underMouse():
                view.autofit = not view.autofit
                break

    def patch_dark_stylesheet(self, stylesheet: str) -> str:
        return stylesheet + 'QGraphicsView { border: 0px; padding: 0px; }'

    def load_script(
        self, script_path: Path, external_args: list[tuple[str, str]] | None = None, reloading: bool = False,
        start_frame: int | None = None
    ) -> None:
        from random import random

        self.external_args = external_args or []

        self.toolbars.playback.stop()
        self.setWindowTitle('VSPreview: %s %s' % (script_path, self.external_args))

        self.statusbar.label.setText('Evaluating')
        self.script_path = script_path

        sys.path.append(str(self.script_path.parent))

        # Rewrite args so external args will be forwarded correctly
        argv_orig = None
        try:
            argv_orig = sys.argv
            sys.argv = [script_path.name]
        except AttributeError:
            pass

        try:
            if reloading:
                std_path_lib = Path(logging.__file__).parent.parent
                std_path_dlls = std_path_lib.parent / 'DLLs'

                check_reloaded = set[str]()

                for module in set(sys.modules.values()) - PRELOADED_MODULES:
                    if not hasattr(module, '__file__') or module.__file__ is None:
                        continue

                    main_mod = module.__name__.split('.')[0]

                    if main_mod in check_reloaded:
                        continue

                    mod_file = Path(module.__file__)

                    if 'vspreview' in mod_file.parts:
                        continue

                    if std_path_lib in mod_file.parents or std_path_dlls in mod_file.parents:
                        continue

                    if not mod_file.exists() or not mod_file.is_file():
                        continue

                    check_reloaded.add(main_mod)

                for module in check_reloaded:
                    all_submodules = sorted([
                        k for k in sys.modules.keys()
                        if k == module or k.startswith(f'{module}.')
                    ])

                    for mod_name in all_submodules:
                        try:
                            if Path(sys.modules[mod_name].__file__).stat().st_mtime > self.last_reload_time:
                                break
                        except Exception:
                            ...
                    else:
                        continue

                    try:
                        logging.info(f'Hot reloaded Python Package: "{module}"')
                        for mod_name in reversed(all_submodules):
                            sys.modules[mod_name] = reload_module(sys.modules[mod_name])
                    except Exception as e:
                        logging.error(e)

            self.env = vpy.variables(
                dict(self.external_args),
                environment=vs.get_current_environment(),
                module_name="__vspreview__"
            ).result()
            self.env.module.__dict__['_monkey_runpy'] = random()
            self.env = vpy.script(self.script_path, environment=self.env).result()
        except vpy.ExecutionFailed as e:
            from traceback import TracebackException

            logging.error(e.parent_error)

            te = TracebackException.from_exception(e.parent_error)
            logging.error(''.join(te.format()))

            self.script_exec_failed = True
            return self.handle_script_error(
                '\n'.join([
                    str(e), 'See console output for details.'
                ]), True
            )
        finally:
            if argv_orig is not None:
                sys.argv = argv_orig
            sys.path.pop()
            self.last_reload_time = time()

        if len(vs.get_outputs()) == 0:
            logging.error('Script has no outputs set.')
            self.script_exec_failed = True
            self.handle_script_error('Script has no outputs set.', True)
            return

        reload_from_error = self.script_exec_failed and reloading
        self.script_exec_failed = False
        self.current_storage_path = (self.current_config_dir / self.script_path.stem).with_suffix('.yml')

        self.storage_not_found = not (
            self.current_storage_path.exists() and self.current_storage_path.read_text('utf8').strip()
        )

        load_error = None

        try:
            if self.storage_not_found or reload_from_error:
                self.load_storage()

            if not reloading:
                self.toolbars.main.rescan_outputs()
                self.toolbars.playback.rescan_outputs()

            if not self.storage_not_found:
                self.load_storage()
        except Exception as e:
            load_error = e

        with self.env:
            vs.register_on_destroy(self.gc_collect)

        if load_error is None:
            self.autosave_timer.start(round(float(self.settings.autosave_interval) * 1000))

            if not reloading:
                self.switch_output(self.settings.output_index)
                if start_frame is not None:
                    self.switch_frame(Frame(start_frame))
        else:
            error_string = "There was an error while loading the script!\n"

            logging.error(error_string + vpy.textwrap.indent(vpy.ExecutionFailed.extract_traceback(load_error), '| '))

            self.script_exec_failed = True

            return self.handle_script_error(
                f'{error_string}{vpy.textwrap.indent(str(load_error), " | ")}\nSee console output for details.', False
            )

    @set_status_label('Loading...')
    def load_storage(self) -> None:
        storage_paths = [self.global_storage_path, self.current_storage_path]

        if self.storage_not_found:
            logging.info('No storage found. Using defaults.')

            if not self.global_storage_path.exists():
                return

            storage_paths = storage_paths[:1]

        storage_contents = ''
        broken_storage = False
        global_length = 0
        for i, storage_path in enumerate(storage_paths):
            try:
                with io.open(storage_path, 'r', encoding='utf-8') as storage_file:
                    version = storage_file.readline()
                    if 'Version' not in version or any({
                        version.strip().endswith(f'@{v}') for v in self.BREAKING_CHANGES_VERSIONS
                    }):
                        raise FileNotFoundError

                    storage_contents += storage_file.read()
                    storage_contents += '\n'

                    if i == 0:
                        global_length = storage_contents.count('\n')
            except FileNotFoundError:
                if self.settings.force_old_storages_removal or i == 0:
                    if storage_path.exists():
                        storage_path.unlink()
                        broken_storage = True
                else:
                    logging.warning(
                        '\n\tThe storage was created on an old version of VSPreview.'
                        '\n\tSave any scening or other important info and delete it.'
                        '\n\tIf you want the program to silently delete old storages, go into settings.'
                    )
                    sys.exit(1)

        if broken_storage:
            return

        loader = yaml_Loader(storage_contents)
        try:
            loader.get_single_data()
        except YAMLError as exc:
            if isinstance(exc, MarkedYAMLError):
                if exc.problem_mark:
                    line = exc.problem_mark.line + 1
                    isglobal = line <= global_length
                    if not isglobal:
                        line -= global_length
                    logging.warning(
                        'Storage ({}) parsing failed on line {} column {}. \n({})\n Exiting...'
                        .format(
                            'Global' if isglobal else 'Local',
                            line, exc.problem_mark.column + 1,
                            str(exc).partition('in "<unicode string>"')[0].strip()
                        )
                    )
                    sys.exit(1)
            else:
                logging.warning('Storage parsing failed. Using defaults.')
        finally:
            loader.dispose()

        if self.settings.color_management:
            assert self.app
            self.current_screen = self.app.primaryScreen()
            self.update_display_profile()

    @fire_and_forget
    @set_status_label('Saving...')
    def dump_storage_async(self) -> None:
        self.dump_storage()

    def dump_storage(self, manually: bool = False) -> None:
        from itertools import count

        if self.script_exec_failed:
            return

        self.current_config_dir.mkdir(0o777, True, True)
        self.global_config_dir.mkdir(0o777, True, True)

        backup_paths = [
            self.current_storage_path.with_suffix(f'.old{i}.yml')
            for i in range(self.settings.STORAGE_BACKUPS_COUNT, 0, -1)
        ] + [self.current_storage_path]

        for src_path, dest_path in zip(backup_paths[1:], backup_paths[:-1]):
            if src_path.exists():
                src_path.replace(dest_path)

        storage_dump = self._dump_serialize(self._serialize_data()).splitlines()

        idx = next(idx for (line, idx) in zip(storage_dump[2:], count(2)) if not line.startswith(' '))

        version = f'# Version@{self.VSP_VERSION}'

        with io.open(self.global_storage_path, 'w', encoding='utf-8') as global_file:
            global_file.writelines(
                '\n'.join([version, '# Global VSPreview storage for settings'] + storage_dump[:idx])
            )

        with io.open(self.current_storage_path, 'w', encoding='utf-8') as current_file:
            current_file.writelines(
                '\n'.join([
                    version,
                    f'# VSPreview local storage for script: {self.script_path}',
                    f'# Global setting (storage/plugins) saved at path: {self.global_config_dir}'
                ] + storage_dump[idx:])
            )

        if manually:
            self.show_message('Saved successfully')

    def _serialize_data(self) -> Any:
        # idk how to explain how this work,
        # but i'm referencing settings objects before in the dict
        # so the yaml serializer will reference the same objects after (in toolbars),
        # which really are the original objects, to those copied in _globals :poppo:
        data = cast(dict[str, Any], self.__getstate__())
        data['_globals'] = {
            'settings': data['settings'],
            'window_settings': data['window_settings']
        }

        data['_globals']['toolbars'] = data['toolbars'].__getstate__()
        gtoolbars = data['_globals']['toolbars']

        for toolbar_name in gtoolbars:
            gtoolbars[toolbar_name].clear()
            gtoolbars[toolbar_name] = getattr(data['toolbars'], toolbar_name).settings

        data['_toolbars_settings'] = [None] * (len(gtoolbars) + 1)
        for i, toolbar_name in enumerate(gtoolbars.keys()):
            data['_toolbars_settings'][i] = gtoolbars[toolbar_name]
        data['_toolbars_settings'][-1] = data['_globals']['window_settings']

        return data

    def _dump_serialize(self, data: Any) -> str:
        storage_dump = io.StringIO()

        dumper = yaml_Dumper(
            storage_dump, default_style=None, default_flow_style=False,
            canonical=None, indent=4, width=120, allow_unicode=True,
            line_break='\n', encoding='utf-8', version=None, tags=None,
            explicit_start=None, explicit_end=None, sort_keys=True
        )
        try:
            dumper.open()
            dumper.represent(data)
            dumper.close()
        finally:
            dumper.dispose()

        return storage_dump.getvalue()

    def init_outputs(self) -> None:
        if not self.outputs:
            return

        self.plugins.init_outputs()

        for graphics_view in self.graphics_views:
            graphics_view.graphics_scene.init_scenes()

    def clean_core_references(self) -> None:
        from vstools.utils.vs_proxy import clear_cache

        for graphics_view in self.graphics_views:
            graphics_view.graphics_scene.clear()

        self.timecodes.clear()
        self.norm_timecodes.clear()

        self.toolbars.pipette._curr_frame_cache.clear()
        self.toolbars.pipette._curr_alphaframe_cache.clear()
        self.toolbars.pipette.outputs.clear()

        for v in self.user_output_info.values():
            for k in v.values():
                k.clear()
            v.clear()

        try:
            with self.env:
                clear_cache()
        except Exception:
            ...

        vs.clear_outputs()

        if self.outputs:
            self.outputs.clear()

        self.gc_collect()

    def reload_script(self) -> None:
        self.reload_before_signal.emit()

        self.dump_storage()

        self.clean_core_references()

        old_environment = get_current_environment()

        self.clear_monkey_runpy()
        make_environment()
        dispose_environment(old_environment)
        self.gc_collect()

        try:
            self.load_script(self.script_path, reloading=True)
        finally:
            self.clear_monkey_runpy()

        self.reload_after_signal.emit()

        self.show_message('Reloaded successfully')

    def clear_monkey_runpy(self) -> None:
        if self.env and '_monkey_runpy' in self.env.module.__dict__:
            key = self.env.module.__dict__['_monkey_runpy']

            if key in _monkey_runpy_dicts:
                _monkey_runpy_dicts[key].clear()
                _monkey_runpy_dicts.pop(key, None)
            elif _monkey_runpy_dicts:
                for env in _monkey_runpy_dicts.values():
                    env.clear()
                _monkey_runpy_dicts.clear()

        self.gc_collect()

    def gc_collect(self) -> None:
        import gc

        for i in range(3):
            gc.collect(generation=i)

        for _ in range(3):
            gc.collect()

    def switch_frame(
        self, pos: Frame | int, *, render_frame: bool | Iterable[vs.VideoFrame | None] = True
    ) -> None:
        frame = Frame(pos)

        if (not 0 <= frame < self.current_output.total_frames):
            return

        if render_frame:
            if isinstance(render_frame, bool):
                self.current_output.render_frame(frame, output_colorspace=self.display_profile)
            else:
                self.current_output.render_frame(
                    frame, *render_frame, output_colorspace=self.display_profile  # type: ignore
                )

        self.current_output.last_showed_frame = frame

        self.timeline.set_position(frame)

        for toolbar in self.toolbars:
            toolbar.on_current_frame_changed(frame)

        self.plugins.on_current_frame_changed(frame)

        self.statusbar.frame_props_label.setText(self.STATUS_FRAME_PROP(self.current_output.props))

    def switch_output(self, value: int | VideoOutput) -> None:
        if not self.outputs or len(self.outputs) == 0:
            return

        if isinstance(value, VideoOutput):
            index = self.outputs.index_of(value)
        else:
            index = value

        if index < 0:
            index = len(self.outputs) + index

        if index < 0 or index >= len(self.outputs):
            return

        prev_index = self.toolbars.main.outputs_combobox.currentIndex()

        self.toolbars.playback.stop()

        # current_output relies on outputs_combobox
        self.toolbars.main.on_current_output_changed(index, prev_index)
        self.timeline.set_end_frame(self.current_output)

        self.switch_frame(self.current_output.last_showed_frame)

        for graphics_view in self.graphics_views:
            for item in graphics_view.graphics_scene.graphics_items:
                item.hide()

            graphics_view.current_scene.show()
            graphics_view.graphics_scene.setSceneRect(
                QRectF(graphics_view.current_scene.pixmap().rect())
            )

        self.timeline.update_notches()

        for toolbar in self.toolbars[1:]:
            toolbar.on_current_output_changed(index, prev_index)

        self.plugins.on_current_output_changed(index, prev_index)

        self.update_statusbar_output_info()

    @property
    def current_output(self) -> VideoOutput:
        return cast(VideoOutput, self.toolbars.main.outputs_combobox.currentData())

    @property
    def outputs(self) -> VideoOutputs | None:
        return self.toolbars.main.outputs

    @property
    def current_scene(self) -> GraphicsImageItem:
        return self.graphics_view.current_scene

    @property
    def graphics_views(self) -> list[GraphicsView]:
        return list(self.bound_graphics_views.keys())

    def register_graphic_view(self, view: GraphicsView) -> None:
        self.bound_graphics_views[view] = {view}

        view.zoom_combobox.currentTextChanged.connect(partial(self.on_zoom_changed, bound_view=view))

        view.zoom_combobox.setModel(GeneralModel[float](self.settings.zoom_levels))
        view.zoom_combobox.setCurrentIndex(self.settings.zoom_default_index)

    def on_zoom_changed(self, text: str | None = None, bound_view: GraphicsView | None = None) -> None:
        if not bound_view:
            return

        for view in self.bound_graphics_views[bound_view]:
            view.setZoom(bound_view.zoom_combobox.currentData())

    def handle_script_error(self, message: str, script: bool = False) -> None:
        self.clear_monkey_runpy()
        self.script_error_dialog.label.setText(message)
        self.script_error_dialog.setWindowTitle('Script Loading Error' if script else 'Program Error')
        self.script_error_dialog.open()

    def on_timeline_clicked(self, start: int) -> None:
        if self.toolbars.playback.play_timer.isActive():
            self.toolbars.playback.stop()
            self.switch_frame(start)
            self.toolbars.playback.play()
        else:
            self.switch_frame(start)

    def update_display_profile(self) -> None:
        if sys.platform == 'win32':
            assert self.app and _imagingcms

            screen_name = self.current_screen.name()

            dc = win32gui.CreateDC('DISPLAY', screen_name, None)

            logging.info(f'Changed screen: {screen_name}')

            icc_path = _imagingcms.get_display_profile_win32(dc, 1)

            if icc_path is not None:
                with open(icc_path, 'rb') as icc:
                    self.display_profile = QColorSpace.fromIccProfile(icc.read())

        if hasattr(self, 'current_output') and self.current_output is not None and self.display_profile is not None:
            self.switch_frame(self.current_output.last_showed_frame)

    def show_message(self, message: str) -> None:
        self.statusbar.showMessage(
            message, round(float(self.settings.statusbar_message_timeout) * 1000)
        )

    def update_statusbar_output_info(self, output: VideoOutput | None = None) -> None:
        output = output or self.current_output
        fmt = output.source.clip.format
        assert fmt

        self.statusbar.total_frames_label.setText(f'{output.total_frames} frames ')
        self.statusbar.duration_label.setText(f'{output.total_time} ')
        self.statusbar.resolution_label.setText(f'{output.width}x{output.height} ')
        self.statusbar.pixel_format_label.setText(f'{fmt.name} ')

        if output.got_timecodes:
            times = sorted(set(output.timecodes), reverse=True)

            if len(times) >= 2:
                return self.statusbar.fps_label.setText(
                    f'VFR {",".join(f"{float(fps):.3f}" for fps in times)} fps '
                )

        if output.fps_den != 0:
            return self.statusbar.fps_label.setText(
                f'{output.fps_num}/{output.fps_den} = {output.fps_num / output.fps_den:.3f} fps '
            )

        self.statusbar.fps_label.setText(f'VFR {output.fps_num}/{output.fps_den} fps ')

    def update_timecodes_info(
        self, index: int, timecodes: str | Path | dict[
            tuple[int | None, int | None], float | tuple[int, int] | Fraction
        ] | list[Fraction], den: int | None = None
    ) -> None:
        self.timecodes[index] = (timecodes, den)

    def set_node_info(self, node_type: type, index: int, **kwargs: Any) -> None:
        base = self.user_output_info[node_type]

        if index not in base:
            base[index] = {**kwargs}
        else:
            base[index] |= kwargs

    def event(self, event: QEvent) -> bool:
        if event.type() == QEvent.Type.LayoutRequest:
            self.timeline.full_repaint()

        return super().event(event)

    # misc methods
    def showEvent(self, event: QShowEvent) -> None:
        super().showEvent(event)
        for graphics_view in self.graphics_views:
            graphics_view.setSizePolicy(self.EVENT_POLICY)

        self.main_split.setSizePolicy(self.EVENT_POLICY)

    def closeEvent(self, event: QCloseEvent) -> None:
        if self.settings.autosave_control.value() != Time(seconds=0):
            self.dump_storage_async()

        self.reload_signal.emit()

    def moveEvent(self, _move_event: QMoveEvent) -> None:
        if self.settings.color_management:
            assert self.app
            screen_number = self.app.primaryScreen()
            if self.current_screen != screen_number:
                self.current_screen = screen_number
                self.update_display_profile()

    def refresh_video_outputs(self) -> None:
        if not self.outputs:
            return

        playback_active = self.toolbars.playback.play_timer.isActive()

        if playback_active:
            self.toolbars.playback.stop()

        self.init_outputs()

        self.switch_output(self.toolbars.main.outputs_combobox.currentIndex())

        if playback_active:
            self.toolbars.playback.play()

    def __getstate__(self) -> Mapping[str, Any]:
        return super().__getstate__() | {
            'window_settings': self.window_settings
        }



from __future__ import annotations

from fractions import Fraction
from itertools import count as iter_count
from typing import TYPE_CHECKING, Any, Mapping, cast

import vapoursynth as vs
from PyQt6 import sip
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QColorSpace, QImage, QPainter, QPixmap

from ..abstracts import AbstractYAMLObject, main_window, try_load
from .misc import CroppingInfo, VideoOutputNode
from .units import Frame, Time

if TYPE_CHECKING:
    from vstools import VideoFormatT
    from ..custom.graphicsview import GraphicsImageItem

__all__ = [
    'VideoOutput'
]


class PackingTypeInfo:
    _getid = iter_count()

    def __init__(
        self, name: str, vs_format: VideoFormatT, qt_format: QImage.Format, shuffle: bool, can_playback: bool = True
    ):
        self.id = next(self._getid)
        self.name = name
        self.vs_format = vs.core.get_video_format(vs_format)
        self.qt_format = qt_format
        self.shuffle = shuffle
        self.can_playback = can_playback

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PackingTypeInfo):
            raise NotImplementedError
        return self.id == other.id

    def __hash__(self) -> int:
        return int(self.id)


class PackingType(PackingTypeInfo):
    none_8bit = PackingTypeInfo('none-8bit', vs.RGB24, QImage.Format.Format_BGR30, False, False)
    none_10bit = PackingTypeInfo('none-10bit', vs.RGB30, QImage.Format.Format_BGR30, False, False)
    numpy_8bit = PackingTypeInfo('numpy-8bit', vs.RGB24, QImage.Format.Format_BGR30, True)
    numpy_10bit = PackingTypeInfo('numpy-10bit', vs.RGB30, QImage.Format.Format_BGR30, True)
    libp2p_8bit = PackingTypeInfo('libp2p-8bit', vs.RGB24, QImage.Format.Format_RGB32, False)
    libp2p_10bit = PackingTypeInfo('libp2p-10bit', vs.RGB30, QImage.Format.Format_BGR30, True)
    akarin_8bit = PackingTypeInfo('akarin-8bit', vs.RGB24, QImage.Format.Format_BGR30, False)
    akarin_10bit = PackingTypeInfo('akarin-10bit', vs.RGB30, QImage.Format.Format_BGR30, False)


PACKING_TYPE: PackingTypeInfo = None  # type: ignore


class VideoOutput(AbstractYAMLObject):
    storable_attrs = (
        'name', 'last_showed_frame', 'play_fps', 'crop_values'
    )

    __slots__ = (
        *storable_attrs, 'index', 'width', 'height', 'fps_num', 'fps_den',
        'total_frames', 'total_time',
        'end_frame', 'fps', 'source', 'prepared',
        'main', 'checkerboard', 'props', '_stateset'
    )

    source: VideoOutputNode
    prepared: VideoOutputNode
    name: str
    last_showed_frame: Frame
    crop_values: CroppingInfo
    _stateset: bool

    def clear(self) -> None:
        if self.source:
            del self.source.clip, self.source.alpha
        if self.prepared:
            del self.prepared.clip, self.prepared.alpha
        if self.props:
            self.props.clear()
        del self.source, self.prepared, self.props
        self.source = self.prepared = self.props = None

    def __init__(
        self, vs_output: vs.VideoOutputTuple, index: int, new_storage: bool = False
    ) -> None:
        self.setValue(vs_output, index, new_storage)

    def setValue(
        self, vs_output: vs.VideoOutputTuple, index: int, new_storage: bool = False
    ) -> None:
        self._stateset = not new_storage

        self.main = main_window()

        assert self.main.env

        self.set_fmt_values()

        # runtime attributes
        self.source = VideoOutputNode(vs_output.clip, vs_output.alpha)
        self.prepared = VideoOutputNode(vs_output.clip, vs_output.alpha)

        if self.source.alpha is not None:
            self.prepared.alpha = self.prepare_vs_output(self.source.alpha, True).std.CopyFrameProps(self.source.alpha)

        with self.main.env:
            vs_outputs = list(vs.get_outputs().values())

        self.vs_index = index
        self.index = vs_outputs.index(vs_output) if vs_output in vs_outputs else index

        self.prepared.clip = self.prepare_vs_output(self.source.clip).std.CopyFrameProps(self.source.clip)
        self.width = self.prepared.clip.width
        self.height = self.prepared.clip.height
        self.fps_num = self.prepared.clip.fps.numerator
        self.fps_den = self.prepared.clip.fps.denominator
        self.fps = self.fps_num / self.fps_den
        self.total_frames = Frame(self.prepared.clip.num_frames)
        self.info = self.main.user_output_info[vs.VideoNode].get(self.vs_index, {})  # type: ignore

        self.name = self.info.get('name', 'Video Node %d' % self.vs_index)

        if self.main.outputs is not None:
            self.main.outputs.setData(self.main.outputs.index(self.vs_index), self.name)

        self.props = cast(vs.FrameProps, {})

        if self.source.alpha:
            self.checkerboard = self._generate_checkerboard()

        if not hasattr(self, 'last_showed_frame') or not (0 <= self.last_showed_frame < self.total_frames):
            self.last_showed_frame = Frame(0)

        if index in self.main.timecodes:
            timecodes, tden = self.main.timecodes[index]

            if self.fps_num == 0:
                try:
                    play_fps = self.main.toolbars.playback.get_true_fps(0, self.props, True)
                except Exception:
                    if isinstance(timecodes, list):
                        play_fps = timecodes[self.last_showed_frame]
                    else:
                        play_fps = Fraction(24000, 1001)
            else:
                play_fps = Fraction(self.fps_num, self.fps_den)

            self.play_fps = play_fps

            if timecodes:
                from pathlib import Path

                from vstools import Timecodes

                if not isinstance(timecodes, list):
                    if isinstance(timecodes, (str, Path)):
                        timecodes = Timecodes.from_file(
                            timecodes, self.source.clip.num_frames, tden, func='set_timecodes'
                        ).to_fractions()

                if isinstance(timecodes, dict):
                    norm_timecodes = Timecodes.normalize_range_timecodes(
                        timecodes, self.source.clip.num_frames, play_fps  # type: ignore
                    )
                else:
                    norm_timecodes = timecodes.copy()

                if len(norm_timecodes) != self.source.clip.num_frames:
                    from vstools import FramesLengthError
                    raise FramesLengthError(
                        'set_timecodes', '', 'timecodes file length mismatch with clip\'s length!',
                        reason=dict(timecodes=len(norm_timecodes), clip=self.source.clip.num_frames)
                    )

                self.main.norm_timecodes[index] = norm_timecodes  # type: ignore
                self.play_fps = Fraction(norm_timecodes[self.last_showed_frame])
        elif not hasattr(self, 'play_fps'):
            if self.fps_num == 0 and self._stateset:
                self.play_fps = self.main.toolbars.playback.get_true_fps(
                    self.last_showed_frame.value, self.props
                )
            else:
                self.play_fps = Fraction(self.fps_num, self.fps_den)

        if index in self.main.norm_timecodes:
            norm_timecodes = self.main.norm_timecodes[index]  # type: ignore

            if (vfr := len(set(norm_timecodes)) > 1) or self.fps_num == 0:
                if not self.main.toolbars.playback.fps_variable_checkbox.isChecked():
                    self.main.toolbars.playback.fps_variable_checkbox.setChecked(True)

            self.got_timecodes = vfr
            self.timecodes = norm_timecodes
        else:
            self.got_timecodes = False

        if self.got_timecodes:
            acc = 0.0
            self._timecodes_frame_to_time = [0.0]
            for fps in self.timecodes:
                acc += round(1 / float(fps), 7)
                self._timecodes_frame_to_time.append(round(acc, 3))
            self.total_time = Time(seconds=acc)
        else:
            self.total_time = self.to_time(self.total_frames - Frame(1))

        if not hasattr(self, 'crop_values'):
            self.crop_values = CroppingInfo(0, 0, self.width, self.height, False, False)

    def set_fmt_values(self) -> None:
        import os
        from ctypes import c_char

        global PACKING_TYPE

        if PACKING_TYPE is not None:
            if hasattr(self, '_FRAME_CONV_INFO'):
                return

            self._NORML_FMT = PACKING_TYPE.vs_format
            self._ALPHA_FMT = vs.core.get_video_format(vs.GRAY8)

            nbps, abps = self._NORML_FMT.bits_per_sample, self._ALPHA_FMT.bytes_per_sample
            self._FRAME_CONV_INFO = {
                False: (c_char * nbps, PACKING_TYPE.qt_format),
                True: (c_char * abps, QImage.Format.Format_Alpha8)
            }

            return

        _default_10bits = os.name != 'nt' and QPixmap.defaultDepth() == 30  # type: ignore

        # From fastest to slowest
        if hasattr(vs.core, 'akarin'):
            PACKING_TYPE = PackingType.akarin_10bit if _default_10bits else PackingType.akarin_8bit
        elif hasattr(vs.core, 'libp2p'):
            PACKING_TYPE = PackingType.libp2p_10bit if _default_10bits else PackingType.libp2p_8bit
        else:
            try:
                import numpy  # noqa: F401
                PACKING_TYPE = PackingType.numpy_10bit if _default_10bits else PackingType.numpy_8bit
            except ModuleNotFoundError:
                PACKING_TYPE = PackingType.none_10bit if _default_10bits else PackingType.none_8bit

        self.set_fmt_values()

    def prepare_vs_output(self, clip: vs.VideoNode, is_alpha: bool = False) -> vs.VideoNode:
        from vstools import ChromaLocation, ColorRange, KwargsT, Matrix, Primaries, Transfer, video_heuristics

        assert clip.format

        heuristics = video_heuristics(clip, None)

        resizer_kwargs = KwargsT({
            'format': self._NORML_FMT.id,
            'matrix_in': Matrix.BT709,
            'transfer_in': Transfer.BT709,
            'primaries_in': Primaries.BT709,
            'range_in': ColorRange.LIMITED,
            'chromaloc_in': ChromaLocation.LEFT
        } | heuristics | {
            'dither_type': self.main.toolbars.playback.settings.dither_type
        })

        if clip.format.color_family == vs.RGB:
            del resizer_kwargs['matrix_in']
        elif clip.format.color_family == vs.GRAY:
            clip = clip.std.RemoveFrameProps('_Matrix')

        if isinstance(resizer_kwargs['range_in'], ColorRange):
            resizer_kwargs['range_in'] = resizer_kwargs['range_in'].value_zimg

        assert clip.format

        if is_alpha:
            if clip.format.id == self._ALPHA_FMT.id:
                return clip
            resizer_kwargs['format'] = self._ALPHA_FMT.id
        elif clip.format.id == vs.GRAY32:
            return clip

        clip = clip.resize.Bicubic(**resizer_kwargs)

        if is_alpha:
            return clip

        return self.pack_rgb_clip(clip)

    def pack_rgb_clip(self, clip: vs.VideoNode) -> vs.VideoNode:
        if PACKING_TYPE.shuffle:
            clip = clip.std.ShufflePlanes([2, 1, 0], vs.RGB)

        shift = 2 ** (10 - PACKING_TYPE.vs_format.bits_per_sample)
        r_shift, g_shift, b_shift = (x * shift for x in (1, 0x400, 0x100000))
        high_bits_mask = 0xc0000000

        if PACKING_TYPE in {
            PackingType.none_8bit, PackingType.none_10bit, PackingType.numpy_8bit, PackingType.numpy_10bit
        }:
            blank = vs.core.std.BlankClip(clip, None, None, vs.GRAY32, color=high_bits_mask, keep=True)

            if PACKING_TYPE in {PackingType.none_8bit, PackingType.none_10bit}:
                from functools import partial
                from multiprocessing.pool import ThreadPool

                from vstools import ranges_product

                indices = list(ranges_product(blank.height, blank.width))

                pool = ThreadPool(self.main.settings.usable_cpus_count * 8)

                def _packing_edarray(
                    bfp: vs.video_view, src_r: vs.video_view,
                    src_g: vs.video_view, src_b: vs.video_view,
                    idx: tuple[int, int]
                ) -> None:
                    bfp[idx] += (src_r[idx] * r_shift) + (src_g[idx] * g_shift) + (src_b[idx] * b_shift)

                def _packrgb(n: int, f: list[vs.VideoFrame]) -> vs.VideoFrame:
                    bf = f[0].copy()

                    pool.map_async(partial(_packing_edarray, bf[0], f[1][0], f[1][1], f[1][2]), indices).wait()

                    return bf
            else:
                import numpy as np

                ein_shift = np.array([b_shift, g_shift, r_shift], np.uint32)

                try:
                    import cupy as cp  # type: ignore

                    self.__base_cupy_add = base_add = cp.asarray(blank.get_frame(0).copy()[0]).copy()

                    ein_shift = cp.asarray(ein_shift)

                    def _packrgb(n: int, f: list[vs.VideoFrame]) -> vs.VideoFrame:
                        bf = f[0].copy()

                        cp.asnumpy(cp.add(cp.einsum(
                            'kji,k->ji', cp.asarray(f[1], cp.uint32), ein_shift, optimize='greedy'
                        ), base_add), out=np.asarray(bf[0]))

                        return bf
                except ModuleNotFoundError:
                    from numpy.core._multiarray_umath import c_einsum  # type: ignore
                    from numpy.core.numeric import tensordot

                    self.__base_numpy_add = base_add = np.asarray(blank.get_frame(0).copy()[0]).copy()

                    def _packrgb(n: int, f: list[vs.VideoFrame]) -> vs.VideoFrame:
                        bf = f[0].copy()

                        np.add(
                            c_einsum(
                                'ji->ji', tensordot(np.asarray(f[1], np.uint32), ein_shift, ((0,), (0,)))
                            ), base_add, np.asarray(bf[0])
                        )

                        return bf

            return blank.std.ModifyFrame([blank, clip], _packrgb)

        if PACKING_TYPE in {PackingType.libp2p_8bit, PackingType.libp2p_10bit}:
            return vs.core.libp2p.Pack(clip)

        if PACKING_TYPE in {PackingType.akarin_8bit, PackingType.akarin_10bit}:
            # x, y, z => b, g, r
            # we want a contiguous array, so we put in 0, 10 bits the R, 11 to 20 the G and 21 to 30 the B
            # R stays like it is * shift if it's 8 bits (gets applied to all planes), then G gets shifted
            # by 10 bits, (we multiply by 2 ** 10) and same for B but by 20 bits and it all gets summed
            return vs.core.akarin.Expr(
                clip.std.SplitPlanes(),
                f'z {b_shift} * y {g_shift} * x {r_shift} * + + {high_bits_mask} +', vs.GRAY32, True
            )

        return clip

    def frame_to_qimage(self, frame: vs.VideoFrame, is_alpha: bool = False) -> QImage:
        from ctypes import POINTER
        from ctypes import cast as ccast

        width, height, stride = frame.width, frame.height, frame.get_stride(0)
        point_size, qt_format = self._FRAME_CONV_INFO[is_alpha]

        pointer = cast(
            sip.voidptr, ccast(frame.get_read_ptr(0), POINTER(point_size * stride)).contents
        )

        return QImage(pointer, width, height, stride, qt_format).copy()  # type: ignore

    @property
    def graphics_scene_item(self) -> GraphicsImageItem | None:
        if self.main.graphics_view.graphics_scene.graphics_items:
            try:
                return self.main.graphics_view.graphics_scene.graphics_items[self.index]
            except IndexError:
                ...

        return None

    def update_graphic_item(
        self, pixmap: QPixmap | None = None, crop_values: CroppingInfo | None | bool = None,
        graphics_scene_item: GraphicsImageItem | None = None
    ) -> QPixmap | None:
        from vstools import complex_hash

        old_crop = complex_hash.hash(self.crop_values)

        if isinstance(crop_values, bool):
            self.crop_values.active = crop_values
        elif crop_values is not None:
            self.crop_values = crop_values

        new_crop = complex_hash.hash(self.crop_values)

        if graphics_scene_item:
            graphics_scene_item.setPixmap(pixmap, self.crop_values)

        if old_crop != new_crop:
            self.main.cropValuesChanged.emit(self.crop_values)

        return pixmap

    def render_frame(
        self, frame: Frame | None, vs_frame: vs.VideoFrame | None = None,
        vs_alpha_frame: vs.VideoFrame | None = None,
        graphics_scene_item: GraphicsImageItem | None = None,
        do_painting: bool = True, output_colorspace: QColorSpace | None = None
    ) -> QPixmap:
        if frame is None or not self._stateset:
            return QPixmap()

        frame = min(max(frame, Frame(0)), self.total_frames - 1)

        vs_frame = vs_frame or self.prepared.clip.get_frame(frame.value)

        self.props = cast(vs.FrameProps, vs_frame.props.copy())

        frame_image = self.frame_to_qimage(vs_frame, False)

        if output_colorspace is not None:
            frame_image.setColorSpace(QColorSpace(QColorSpace.NamedColorSpace.SRgb))
            frame_image.convertToColorSpace(output_colorspace)

        if not vs_frame.closed:
            vs_frame.close()
            del vs_frame

        if graphics_scene_item is None:
            graphics_scene_item = self.graphics_scene_item

        if self.prepared.alpha is None:
            qpixmap = QPixmap.fromImage(frame_image, Qt.ImageConversionFlag.NoFormatConversion)

            if do_painting:
                self.update_graphic_item(qpixmap, graphics_scene_item=graphics_scene_item)

            return qpixmap

        alpha_image = self.frame_to_qimage(
            vs_alpha_frame or self.prepared.alpha.get_frame(frame.value), True
        )

        result_image = QImage(frame_image.size(), QImage.Format.Format_ARGB32_Premultiplied)
        painter = QPainter(result_image)
        painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_Source)
        painter.drawImage(0, 0, frame_image)
        painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_DestinationIn)
        painter.drawImage(0, 0, alpha_image)

        if self.main.toolbars.playback.settings.CHECKERBOARD_ENABLED:
            painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_DestinationOver)
            painter.drawImage(0, 0, self.checkerboard)

        painter.end()

        qpixmap = QPixmap.fromImage(result_image, Qt.ImageConversionFlag.NoFormatConversion)

        if do_painting:
            self.update_graphic_item(qpixmap, graphics_scene_item=graphics_scene_item)

        return qpixmap

    def _generate_checkerboard(self) -> QImage:
        tile_size = self.main.toolbars.playback.settings.CHECKERBOARD_TILE_SIZE
        tile_color_1 = self.main.toolbars.playback.settings.CHECKERBOARD_TILE_COLOR_1
        tile_color_2 = self.main.toolbars.playback.settings.CHECKERBOARD_TILE_COLOR_2

        macrotile_pixmap = QPixmap(tile_size * 2, tile_size * 2)
        painter = QPainter(macrotile_pixmap)
        painter.fillRect(macrotile_pixmap.rect(), tile_color_1)
        painter.fillRect(tile_size, 0, tile_size, tile_size, tile_color_2)
        painter.fillRect(0, tile_size, tile_size, tile_size, tile_color_2)
        painter.end()

        result_image = QImage(self.width, self.height, QImage.Format.Format_ARGB32_Premultiplied)
        painter = QPainter(result_image)
        painter.drawTiledPixmap(result_image.rect(), macrotile_pixmap)
        painter.end()

        return result_image

    def _calculate_frame(self, seconds: float) -> int:
        if self.got_timecodes:
            seconds = float(f'{round(seconds, 7):.6f}')

            ref, maxx = int(self.last_showed_frame), int(self.total_frames)
            low, high = max(ref - 6, 0), min(ref + 6, maxx - 1)

            if (
                li := int(self._timecodes_frame_to_time[low] > seconds)
            ) or (
                hi := int(self._timecodes_frame_to_time[high] < seconds)
            ):
                while self._timecodes_frame_to_time[low] > seconds and low > 0:
                    low -= 6 * li
                    li += 1

                while self._timecodes_frame_to_time[high] < seconds and high < maxx:
                    high += 6 * hi
                    hi += 1

                low, high = max(low - 1, 0), min(high + 1, maxx - 1)

            for i, time in zip(range(high, low - 1, -1), reversed(self._timecodes_frame_to_time[low:high + 1])):
                if time == seconds:
                    return i

                if time < seconds:
                    return i + 1

        return round(seconds * self.fps)

    def _calculate_seconds(self, frame_num: int) -> float:
        if self.got_timecodes:
            return self._timecodes_frame_to_time[frame_num]
        return frame_num / (self.fps or 1)

    def to_frame(self, time: Time) -> Frame:
        return Frame(self._calculate_frame(float(time)))

    def to_time(self, frame: Frame) -> Time:
        return Time(seconds=self._calculate_seconds(int(frame)))

    def with_node(self, new_node: vs.VideoNode | VideoOutputNode) -> VideoOutput:
        if isinstance(new_node, vs.VideoNode):
            new_node = VideoOutputNode(new_node, None)

        new_output = VideoOutput(new_node, self.vs_index, False)
        new_output.index = self.index

        new_output.last_showed_frame = self.last_showed_frame
        new_output.name = self.name

        return new_output

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        try_load(state, 'name', str, self.__setattr__)
        try_load(state, 'last_showed_frame', Frame, self.__setattr__)
        try_load(state, 'play_fps', Fraction, self.__setattr__)
        try_load(state, 'crop_values', CroppingInfo, self.__setattr__)

        self._stateset = True



from __future__ import annotations

from typing import Any, Mapping

from PyQt6.QtWidgets import QLabel

from ...core import AbstractToolbarSettings, CheckBox, HBoxLayout, Time, TimeEdit, try_load

__all__ = [
    'BenchmarkSettings'
]


class BenchmarkSettings(AbstractToolbarSettings):
    __slots__ = (
        'clear_cache_checkbox',
        'refresh_interval_control', 'frame_data_sharing_fix_checkbox',
    )

    def setup_ui(self) -> None:
        super().setup_ui()

        self.clear_cache_checkbox = CheckBox('Clear VS frame caches before each run', self)

        self.frame_data_sharing_fix_checkbox = CheckBox('(Debug) Enable frame data sharing fix', self)

        self.refresh_interval_control = TimeEdit(self)

        self.vlayout.addWidgets([
            self.clear_cache_checkbox,
            self.frame_data_sharing_fix_checkbox
        ])
        self.vlayout.addLayout(
            HBoxLayout([
                QLabel('Refresh interval', self),
                self.refresh_interval_control
            ])
        )

    def set_defaults(self) -> None:
        self.clear_cache_checkbox.setChecked(False)
        self.refresh_interval_control.setValue(Time(milliseconds=150))
        self.frame_data_sharing_fix_checkbox.setChecked(True)

    @property
    def clear_cache_enabled(self) -> bool:
        return self.clear_cache_checkbox.isChecked()

    @property
    def refresh_interval(self) -> Time:
        return self.refresh_interval_control.value()

    @property
    def frame_data_sharing_fix_enabled(self) -> bool:
        return self.frame_data_sharing_fix_checkbox.isChecked()

    def __getstate__(self) -> Mapping[str, Any]:
        return {
            'clear_cache_enabled': self.clear_cache_enabled,
            'refresh_interval': self.refresh_interval,
            'frame_data_sharing_fix_enabled': self.frame_data_sharing_fix_enabled,
        }

    def _setstate_(self, state: Mapping[str, Any]) -> None:
        try_load(state, 'clear_cache_enabled', bool, self.clear_cache_checkbox.setChecked)
        try_load(state, 'refresh_interval', Time, self.refresh_interval_control.setValue)
        try_load(state, 'frame_data_sharing_fix_enabled', bool, self.frame_data_sharing_fix_checkbox.setChecked)



from __future__ import annotations

from typing import Any

from ..abstracts import AbstractYAMLObject

__all__ = [
    'YAMLObjectWrapper'
]


class YAMLObjectWrapper(AbstractYAMLObject):
    value: Any

    def __int__(self) -> int:
        return int(self.value)

    def __float__(self) -> float:
        return float(self.value)

    def __index__(self) -> int:
        return int(self)

    def __str__(self) -> str:
        return str(self.value)

    def __eq__(self, other: object) -> bool:
        if other is None:
            return False
        if not isinstance(other, YAMLObjectWrapper):
            other = self.__class__(other)
        return bool(self.value == other.value)

    def __gt__(self, other: object) -> bool:
        if other is None:
            return False
        if not isinstance(other, YAMLObjectWrapper):
            other = self.__class__(other)
        return bool(self.value > other.value)

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __le__(self, other: object) -> bool:
        return not self.__gt__(other)

    def __ge__(self, other: object) -> bool:
        if other is None:
            return False
        if not isinstance(other, YAMLObjectWrapper):
            other = self.__class__(other)
        return self.__eq__(other) or self.__gt__(other)

    def __lt__(self, other: object) -> bool:
        return not self.__ge__(other)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.value})'

    def __get_storable_attr__(self) -> tuple[str, ...]:
        return self.__slots__



from __future__ import annotations

import atexit
import runpy
import sys
from concurrent.futures import Future
from threading import Lock
from typing import Any, Callable, TypeVar

from PyQt6.QtCore import QObject, QRunnable, QThreadPool, pyqtSignal
from vapoursynth import LogHandle
from vsengine.loops import EventLoop, set_loop  # type: ignore[import]
from vsengine.policy import GlobalStore, ManagedEnvironment, Policy  # type: ignore[import]

from .logger import get_vs_logger

__all__ = [
    'PRELOADED_MODULES',

    '_monkey_runpy_dicts',

    'set_vsengine_loop',
    'get_current_environment',
    'make_environment', 'dispose_environment'
]


PRELOADED_MODULES = set(sys.modules.values())

_monkey_runpy_dicts = {}

orig_runpy_run_code = runpy._run_code  # type: ignore


def _monkey_runpy_func(*args: Any, **kwargs: Any) -> Any:
    glob_dict = orig_runpy_run_code(*args, **kwargs)

    if '_monkey_runpy' in glob_dict:
        _monkey_runpy_dicts[glob_dict['_monkey_runpy']] = glob_dict

    return glob_dict


runpy._run_code = _monkey_runpy_func  # type: ignore

T = TypeVar("T")


class Runner(QRunnable):
    def __init__(self, wrapper: Callable[[], None]) -> None:
        super().__init__()
        self.wrapper = wrapper

    def run(self) -> None:
        self.wrapper()


class PyQTLoop(QObject, EventLoop):
    move = pyqtSignal(int)

    def attach(self) -> None:
        self._signal_lock = Lock()
        self._signal_counter = 0
        self._signallers = dict[int, Callable[[], None]]()

        self._slot = self._receive_task
        self.move.connect(self._slot)

    def detach(self) -> None:
        self.move.disconnect(self._slot)
        self._signallers = {}

    def _receive_task(self, number: int) -> None:
        self._signallers.get(number, lambda: None)()

    def from_thread(self, func: Callable[..., T], *args: Any, **kwargs: Any) -> Future[T]:
        fut = Future[T]()

        with self._signal_lock:
            my_counter = self._signal_counter

            self._signal_counter += 1

        def wrapper() -> None:
            nonlocal my_counter

            if not fut.set_running_or_notify_cancel():
                return

            del self._signallers[my_counter]

            try:
                result = func(*args, **kwargs)
            except BaseException as e:
                fut.set_exception(e)
            else:
                fut.set_result(result)

        self._signallers[my_counter] = wrapper
        self.move.emit(my_counter)

        return fut

    def to_thread(self, func: Callable[..., T], *args: Any, **kwargs: Any) -> Future[T]:
        fut = Future[T]()

        def wrapper() -> None:
            if not fut.set_running_or_notify_cancel():
                return

            try:
                result = func(*args, **kwargs)
            except BaseException as e:
                fut.set_exception(e)
            else:
                fut.set_result(result)

        QThreadPool.globalInstance().start(Runner(wrapper))

        return fut


def set_vsengine_loop() -> None:
    set_loop(PyQTLoop())


current_log: LogHandle | None = None


def make_environment() -> None:
    global environment, current_log
    assert policy is not None

    if environment and current_log:
        environment.core.remove_log_handler(current_log)

    environment = policy.new_environment()
    current_log = environment.core.add_log_handler(get_vs_logger())
    environment.switch()


def dispose_environment(env: ManagedEnvironment) -> None:
    import logging

    from vstools.utils.vs_proxy import _finalize_core

    if current_log:
        env.core.remove_log_handler(current_log)

    _finalize_core(env._environment.env_id, id(env.core), True)

    if logging.getLogger().level <= logging.DEBUG:
        from gc import get_referrers

        # There has to be the environment referencing it
        ref_count = len(get_referrers(env.core)) - 1

        if ref_count:
            logging.debug(f'Core {id(env.core)} is being held reference by {ref_count} objects!')

    env.dispose()


policy: Policy = Policy(GlobalStore())
policy.register()

environment: ManagedEnvironment | None = None

make_environment()


def get_current_environment() -> ManagedEnvironment:
    assert environment
    return environment


def _dispose() -> None:
    global current_log

    if environment:
        dispose_environment(environment)

    current_log = None


atexit.register(_dispose)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, Iterator, Mapping, OrderedDict, TypeVar

import vapoursynth as vs
from PyQt6.QtCore import QAbstractListModel, QModelIndex, Qt

from ..core import AudioOutput, QYAMLObject, VideoOutput, main_window, try_load

if TYPE_CHECKING:
    from ..main import MainWindow

T = TypeVar('T', VideoOutput, AudioOutput)


__all__ = [
    'Outputs',
    'VideoOutputs',
    'AudioOutputs'
]


class Outputs(Generic[T], QAbstractListModel, QYAMLObject):
    __slots__ = ('items', )

    out_type: type[T]
    vs_type: type[vs.VideoOutputTuple] | type[vs.AudioNode]

    items: list[T]
    _items: list[T]

    main: MainWindow

    def __init__(self, main: MainWindow, local_storage: Mapping[str, T] | None = None) -> None:
        self.setValue(main, local_storage)

    def setValue(self, main: MainWindow, local_storage: Mapping[str, T] | None = None) -> None:
        super().__init__()

        self.items = []
        self.main = main

        local_storage, newstorage = (local_storage, False) if local_storage is not None else ({}, True)

        if main.storage_not_found:
            newstorage = False

        outputs = OrderedDict(sorted(vs.get_outputs().items()))

        for i, vs_output in outputs.items():
            if not isinstance(vs_output, self.vs_type):
                continue

            try:
                output = local_storage[str(i)]
                output.setValue(vs_output, i, newstorage)  # type: ignore
            except KeyError:
                output = self.out_type(vs_output, i, newstorage)  # type: ignore

            self.items.append(output)

        self._items = list(self.items)

    def clear_outputs(self) -> None:
        for output in (*self.items, *self._items):
            output.clear()

    def __getitem__(self, i: int) -> T:
        return self.items[i]

    def __len__(self) -> int:
        return len(self.items)

    def index_of(self, item: T) -> int:
        return self.items.index(item)

    def __iter__(self) -> Iterator[T]:
        return iter(self.items)

    def append(self, item: T) -> int:
        index = len(self.items)
        self.beginInsertRows(QModelIndex(), index, index)
        self.items.append(item)
        self.endInsertRows()

        return len(self.items) - 1

    def clear(self) -> None:
        self.clear_outputs()
        self.beginRemoveRows(QModelIndex(), 0, len(self.items))
        self.items.clear()
        self._items.clear()
        self.endRemoveRows()

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.UserRole) -> Any:
        if not index.isValid():
            return None
        if index.row() >= len(self.items):
            return None

        if role == Qt.ItemDataRole.DisplayRole:
            return self.items[index.row()].name
        if role == Qt.ItemDataRole.EditRole:
            return self.items[index.row()].name
        if role == Qt.ItemDataRole.UserRole:
            return self.items[index.row()]
        return None

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self.items)

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemFlag.ItemIsEnabled

        return super().flags(index) | Qt.ItemFlag.ItemIsEditable

    def setData(self, index: QModelIndex, value: Any, role: int = Qt.ItemDataRole.EditRole) -> bool:
        if not index.isValid():
            return False
        if not role == Qt.ItemDataRole.EditRole:
            return False
        if not isinstance(value, str):
            return False

        self.items[index.row()].name = value
        self.dataChanged.emit(index, index, [role])

        return True

    def __getstate__(self) -> Mapping[str, Any]:
        return dict(zip([str(x.index) for x in self.items], self.items), type=self.out_type.__name__)

    def __setstate__(self, state: Mapping[str, T]) -> None:
        type_string = ''
        try_load(state, 'type', str, type_string)

        for key, value in state.items():
            if key == 'type':
                continue
            if not isinstance(key, str):
                raise TypeError(f'Storage loading (Outputs): key {key} is not a string')
            if not isinstance(value, self.out_type):
                raise TypeError(f'Storage loading (Outputs): value of key {key} is not {self.out_type.__name__}')

        self.setValue(main_window(), state)


class VideoOutputs(Outputs[VideoOutput]):
    out_type = VideoOutput
    vs_type = vs.VideoOutputTuple


class AudioOutputs(Outputs[AudioOutput]):
    out_type = AudioOutput
    vs_type = vs.AudioNode  # type: ignore



from __future__ import annotations

from typing import Any, Mapping

from PyQt6.QtWidgets import QLabel

from ...core import AbstractToolbarSettings, CheckBox, HBoxLayout, LineEdit, VBoxLayout, try_load

__all__ = [
    'CompSettings'
]


class CompSettings(AbstractToolbarSettings):
    __slots__ = ('delete_cache_checkbox', 'login_browser_id', 'login_session')

    DEFAULT_COLLECTION_NAME = 'Unknown'

    def setup_ui(self) -> None:
        super().setup_ui()

        self.delete_cache_checkbox = CheckBox('Delete images cache after upload')

        self.login_browser_id = LineEdit('Browser ID')
        self.login_session = LineEdit('Session ID')

        label = QLabel(
            'To get this info: Open Dev console in browser, go to network tab, upload a comparison,'
            'click request called "comparison" Copy browserId from payload, copy session token from '
            'SLPSESSION cookie from cookies'
        )
        label.setMaximumHeight(50)
        label.setMinimumWidth(400)
        label.setWordWrap(True)

        self.vlayout.addWidget(self.delete_cache_checkbox)

        HBoxLayout(
            self.vlayout,
            VBoxLayout([
                label,
                HBoxLayout([QLabel("Browser ID"), self.login_browser_id]),
                HBoxLayout([QLabel("Session ID"), self.login_session])
            ])
        )

    def set_defaults(self) -> None:
        self.delete_cache_checkbox.setChecked(True)

    @property
    def delete_cache_enabled(self) -> bool:
        return self.delete_cache_checkbox.isChecked()

    @property
    def browser_id(self) -> str:
        return self.login_browser_id.text()

    @property
    def session_id(self) -> str:
        return self.login_session.text()

    def __getstate__(self) -> Mapping[str, Any]:
        return {
            'delete_cache_enabled': self.delete_cache_enabled,
            'browser_id': self.browser_id,
            'session_id': self.session_id,
        }

    def _setstate_(self, state: Mapping[str, Any]) -> None:
        try_load(state, 'delete_cache_enabled', bool, self.delete_cache_checkbox.setChecked)
        try_load(state, 'browser_id', str, self.login_browser_id.setText)
        try_load(state, 'session_id', str, self.login_session.setText)



from __future__ import annotations

import sys
from functools import partial
from multiprocessing import cpu_count
from typing import Any, Mapping, cast

from PyQt6.QtCore import QKeyCombination, Qt
from PyQt6.QtGui import QShortcut
from PyQt6.QtWidgets import QComboBox, QLabel

from ..core import (
    AbstractToolbarSettings, CheckBox, ComboBox, HBoxLayout, PushButton, QYAMLObjectSingleton, SpinBox, Time, TimeEdit,
    VBoxLayout, main_window, try_load
)
from ..models import GeneralModel

__all__ = [
    'MainSettings',
    'WindowSettings'
]


class MainSettings(AbstractToolbarSettings):
    __slots__ = (
        'autosave_control', 'base_ppi_spinbox', 'dark_theme_checkbox',
        'opengl_rendering_checkbox', 'output_index_spinbox',
        'png_compressing_spinbox', 'statusbar_timeout_control',
        'timeline_notches_margin_spinbox', 'usable_cpus_spinbox',
        'zoom_levels_combobox', 'zoom_levels_lineedit', 'zoom_level_default_combobox',
        'azerty_keyboard_checkbox', 'dragnavigator_timeout_spinbox', 'color_management_checkbox'
    )

    INSTANT_FRAME_UPDATE = False
    SYNC_OUTPUTS = True
    STORAGE_BACKUPS_COUNT = 2

    def setup_ui(self) -> None:
        super().setup_ui()

        self.autosave_control = TimeEdit(self)

        self.base_ppi_spinbox = SpinBox(self, 1, 999)

        self.dark_theme_checkbox = CheckBox('Dark theme', self)

        self.opengl_rendering_checkbox = CheckBox('OpenGL rendering', self)

        self.force_old_storages_removal_checkbox = CheckBox('Remove old storages', self)

        self.output_index_spinbox = SpinBox(self, 0, 65535)

        self.png_compressing_spinbox = SpinBox(self, 0, 100)

        self.statusbar_timeout_control = TimeEdit(self)

        self.timeline_notches_margin_spinbox = SpinBox(self, 1, 9999, '%')

        self.usable_cpus_spinbox = SpinBox(self, 1, self.get_usable_cpus_count())

        self.azerty_keyboard_checkbox = CheckBox('AZERTY Keyboard', self)

        self.zoom_levels_combobox = ComboBox[int](editable=True, insertPolicy=QComboBox.InsertPolicy.NoInsert)
        self.zoom_levels_lineedit = self.zoom_levels_combobox.lineEdit()

        self.zoom_levels_lineedit.returnPressed.connect(self.zoom_levels_combobox_on_add)
        QShortcut(  # type: ignore
            QKeyCombination(Qt.Modifier.CTRL, Qt.Key.Key_Delete).toCombined(), self.zoom_levels_combobox,
            activated=partial(self.zoom_levels_combobox_on_remove, True)
        )

        self.zoom_level_default_combobox = ComboBox[int]()

        self.dragnavigator_timeout_spinbox = SpinBox(self, 0, 1000 * 60 * 5)

        self.color_management_checkbox = CheckBox('Color management', self)

        HBoxLayout(self.vlayout, [QLabel('Autosave interval (0 - disable)'), self.autosave_control])

        HBoxLayout(self.vlayout, [QLabel('Base PPI'), self.base_ppi_spinbox])

        HBoxLayout(self.vlayout, [
            VBoxLayout([self.dark_theme_checkbox, self.opengl_rendering_checkbox]),
            VBoxLayout([self.force_old_storages_removal_checkbox, self.azerty_keyboard_checkbox])
        ])

        HBoxLayout(self.vlayout, [QLabel('Default output index'), self.output_index_spinbox])

        HBoxLayout(self.vlayout, [QLabel('PNG compression level (0 for max)'), self.png_compressing_spinbox])

        HBoxLayout(self.vlayout, [QLabel('Status bar message timeout'), self.statusbar_timeout_control])

        HBoxLayout(self.vlayout, [
            QLabel('Timeline label notches margin', self), self.timeline_notches_margin_spinbox
        ])

        HBoxLayout(self.vlayout, [QLabel('Usable CPUs count'), self.usable_cpus_spinbox])

        HBoxLayout(self.vlayout, [
            VBoxLayout([
                QLabel('Zoom Levels'),
                HBoxLayout([
                    self.zoom_levels_combobox,
                    PushButton('❌', clicked=self.zoom_levels_combobox_on_remove, maximumWidth=18),
                    PushButton('✔️', clicked=self.zoom_levels_combobox_on_add, maximumWidth=18),
                ])
            ]),
            VBoxLayout([QLabel('Default Zoom Level'), self.zoom_level_default_combobox])
        ])

        HBoxLayout(self.vlayout, [QLabel('Drag Navigator Timeout (ms)'), self.dragnavigator_timeout_spinbox])

        if sys.platform == 'win32':
            HBoxLayout(self.vlayout, [self.color_management_checkbox])

    def set_defaults(self) -> None:
        self.autosave_control.setValue(Time(seconds=30))
        self.base_ppi_spinbox.setValue(96)
        self.dark_theme_checkbox.setChecked(True)
        self.opengl_rendering_checkbox.setChecked(False)
        self.output_index_spinbox.setValue(0)
        self.png_compressing_spinbox.setValue(0)
        self.statusbar_timeout_control.setValue(Time(seconds=2.5))
        self.timeline_notches_margin_spinbox.setValue(20)
        self.force_old_storages_removal_checkbox.setChecked(False)
        self.azerty_keyboard_checkbox.setChecked(False)
        self.usable_cpus_spinbox.setValue(self.get_usable_cpus_count())
        self.dragnavigator_timeout_spinbox.setValue(250)

        self.zoom_levels = [
            25, 50, 68, 75, 85, 100, 150, 200, 400, 600, 800, 1000, 1200, 1400, 1600, 2000, 3200
        ]
        self.zoom_level_default_combobox.setCurrentIndex(5)
        self.color_management_checkbox.setChecked(False)

    @property
    def autosave_interval(self) -> Time:
        return self.autosave_control.value()

    @property
    def base_ppi(self) -> int:
        return self.base_ppi_spinbox.value()

    @property
    def dark_theme_enabled(self) -> bool:
        return self.dark_theme_checkbox.isChecked()

    @property
    def opengl_rendering_enabled(self) -> bool:
        return self.opengl_rendering_checkbox.isChecked()

    @property
    def output_index(self) -> int:
        return self.output_index_spinbox.value()

    @property
    def png_compression_level(self) -> int:
        return self.png_compressing_spinbox.value()

    @property
    def statusbar_message_timeout(self) -> Time:
        return self.statusbar_timeout_control.value()

    @property
    def timeline_label_notches_margin(self) -> int:
        return self.timeline_notches_margin_spinbox.value()

    @property
    def force_old_storages_removal(self) -> int:
        return self.force_old_storages_removal_checkbox.isChecked()

    @property
    def azerty_keybinds(self) -> int:
        return self.azerty_keyboard_checkbox.isChecked()

    @property
    def usable_cpus_count(self) -> int:
        return self.usable_cpus_spinbox.value()

    @property
    def zoom_levels(self) -> list[float]:
        return [
            int(self.zoom_levels_combobox.itemText(i)) / 100
            for i in range(self.zoom_levels_combobox.count())
        ]

    @zoom_levels.setter
    def zoom_levels(self, new_levels: list[int]) -> None:
        new_levels = sorted(set(map(int, new_levels)))

        if len(new_levels) < 3:
            return

        old_values = [int(x * 100) for x in self.zoom_levels]

        self.zoom_levels_combobox.clear()
        self.zoom_levels_combobox.addItems(map(str, new_levels))

        old_default = self.zoom_level_default_combobox.currentData()

        self.zoom_level_default_combobox.setModel(GeneralModel[int](new_levels))

        if old_default:
            try:
                old_default_idx = new_levels.index(int(old_default))
            except ValueError:
                old_default_idx = old_values.index(int(old_default))
                old_default_idx = min(max(old_default_idx - 1, 0), old_default_idx + 1, len(new_levels) - 1)
            self.zoom_level_default_combobox.setCurrentIndex(old_default_idx)

        if hasattr((main := main_window()), 'graphics_view'):
            main_zoom_comb = main.graphics_view.zoom_combobox
            old_index = main_zoom_comb.currentIndex()
            main_zoom_comb.setModel(GeneralModel[float](self.zoom_levels))
            main_zoom_comb.setCurrentIndex(min(max(old_index, 0), len(new_levels) - 1))

        self.zoom_levels_lineedit.clear()

    @property
    def zoom_default_index(self) -> int:
        return self.zoom_level_default_combobox.currentIndex()

    @staticmethod
    def get_usable_cpus_count() -> int:
        from os import getpid
        try:
            from win32.win32api import OpenProcess  # type: ignore
            from win32.win32process import GetProcessAffinityMask  # type: ignore
            from win32con import PROCESS_QUERY_LIMITED_INFORMATION  # type: ignore
            proc_mask, _ = GetProcessAffinityMask(OpenProcess(PROCESS_QUERY_LIMITED_INFORMATION, False, getpid()))
            cpus = [i for i in range(64) if (1 << i) & proc_mask]
            return len(cpus)
        except Exception:
            try:
                from os import sched_getaffinity  # type: ignore
                return len(sched_getaffinity(getpid()))
            except Exception:
                return cpu_count()

    @property
    def dragnavigator_timeout(self) -> int:
        return self.dragnavigator_timeout_spinbox.value()

    def zoom_levels_combobox_on_add(self) -> None:
        try:
            new_value = int(self.zoom_levels_lineedit.text())
        except ValueError:
            return

        if not new_value:
            return

        zoom_levels = [x * 100 for x in self.zoom_levels]

        if new_value in zoom_levels:
            return

        self.zoom_levels = [*zoom_levels, new_value]

    def zoom_levels_combobox_on_remove(self, checkFocus: bool = False) -> None:
        if checkFocus and not self.zoom_levels_lineedit.hasFocus():
            return

        try:
            old_value = int(self.zoom_levels_lineedit.text())
        except ValueError:
            return

        if not old_value:
            return

        zoom_levels = [x * 100 for x in self.zoom_levels]

        if old_value not in zoom_levels:
            return

        self.zoom_levels = [x for x in zoom_levels if round(x) != round(old_value)]

    @property
    def color_management(self) -> bool:
        return self.color_management_checkbox.isChecked()

    def __getstate__(self) -> Mapping[str, Any]:
        return {
            'autosave_interval': self.autosave_interval,
            'base_ppi': self.base_ppi,
            'dark_theme': self.dark_theme_enabled,
            'opengl_rendering': self.opengl_rendering_enabled,
            'output_index': self.output_index,
            'png_compression': self.png_compression_level,
            'statusbar_message_timeout': self.statusbar_message_timeout,
            'timeline_label_notches_margin': self.timeline_label_notches_margin,
            'force_old_storages_removal': self.force_old_storages_removal,
            'zoom_levels': sorted([int(x * 100) for x in self.zoom_levels]),
            'zoom_default_index': self.zoom_default_index,
            'dragnavigator_timeout': self.dragnavigator_timeout,
            'color_management': self.color_management
        }

    def _setstate_(self, state: Mapping[str, Any]) -> None:
        try_load(state, 'autosave_interval', Time, self.autosave_control.setValue)
        try_load(state, 'base_ppi', int, self.base_ppi_spinbox.setValue)
        try_load(state, 'dark_theme', bool, self.dark_theme_checkbox.setChecked)
        try_load(state, 'opengl_rendering', bool, self.opengl_rendering_checkbox.setChecked)
        try_load(state, 'output_index', int, self.output_index_spinbox.setValue)
        try_load(state, 'png_compression', int, self.png_compressing_spinbox.setValue)
        try_load(state, 'statusbar_message_timeout', Time, self.statusbar_timeout_control.setValue)
        try_load(state, 'timeline_label_notches_margin', int, self.timeline_notches_margin_spinbox.setValue)
        try_load(state, 'force_old_storages_removal', bool, self.force_old_storages_removal_checkbox.setChecked)
        try_load(state, 'zoom_levels', list, self)
        try_load(state, 'zoom_default_index', int, self.zoom_level_default_combobox.setCurrentIndex)
        try_load(state, 'dragnavigator_timeout', int, self.dragnavigator_timeout_spinbox.setValue)
        try_load(state, 'color_management', bool, self.color_management_checkbox.setChecked)


class WindowSettings(QYAMLObjectSingleton):
    __slots__ = (
        'timeline_mode', 'window_geometry', 'window_state', 'zoom_index', 'x_pos', 'y_pos'
    )

    def __getstate__(self) -> Mapping[str, Any]:
        main = main_window()

        return {
            'timeline_mode': main.timeline.mode,
            'window_geometry': bytes(cast(bytearray, main.saveGeometry())),
            'window_state': bytes(cast(bytearray, main.saveState())),
            'zoom_index': main.graphics_view.zoom_combobox.currentIndex(),
            'x_pos': main.graphics_view.horizontalScrollBar().value(),
            'y_pos': main.graphics_view.verticalScrollBar().value(),
        }

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        try_load(state, 'timeline_mode', str, self.__setattr__)
        try_load(state, 'window_geometry', bytes, self.__setattr__)
        try_load(state, 'window_state', bytes, self.__setattr__)
        try_load(state, 'zoom_index', int, self.__setattr__)
        try_load(state, 'x_pos', int, self.__setattr__)
        try_load(state, 'y_pos', int, self.__setattr__)

        main = main_window()

        main.timeline.mode = self.timeline_mode

        main.graphics_view.zoom_combobox.setCurrentIndex(self.zoom_index)

        main.graphics_view.horizontalScrollBar().setValue(self.x_pos)
        main.graphics_view.verticalScrollBar().setValue(self.y_pos)

        main.restoreState(self.window_state)
        main.restoreGeometry(self.window_geometry)



from __future__ import annotations

from datetime import timedelta
from typing import Any, Mapping, SupportsFloat, SupportsInt, Union, cast

from ..abstracts import main_window, try_load
from .yaml import YAMLObjectWrapper

__all__ = [
    'Frame', 'Time'
]


Number = Union[int, float, SupportsInt, SupportsFloat]


class Frame(YAMLObjectWrapper):
    __slots__ = ('value',)

    def __init__(self, init_value: Number | Frame | Time | None = 0) -> None:
        if isinstance(init_value, float):
            init_value = int(init_value)
        if isinstance(init_value, int):
            self.value = init_value
        elif isinstance(init_value, Frame):
            self.value = init_value.value
        elif isinstance(init_value, Time):
            self.value = main_window().current_output.to_frame(init_value).value
        else:
            raise TypeError

    def __add__(self, other: Number | Frame) -> Frame:
        if not isinstance(other, Frame):
            other = Frame(other)
        return Frame(self.value + other.value)

    def __iadd__(self, other: Number | Frame) -> Frame:
        if not isinstance(other, Frame):
            other = Frame(other)
        self.value += other.value
        return self

    def __sub__(self, other: Number | Frame) -> Frame:
        if isinstance(other, Frame):
            return Frame(self.value - other.value)
        return self - Frame(other)

    def __isub__(self, other: Number | Frame) -> Frame:
        if not isinstance(other, Frame):
            other = Frame(other)
        self.value -= other.value
        return self

    def __mul__(self, other: Number | Frame) -> Frame:
        if not isinstance(other, Frame):
            other = Frame(other)
        return Frame(self.value * other.value)

    def __imul__(self, other: Number | Frame) -> Frame:
        if not isinstance(other, Frame):
            other = Frame(other)
        self.value *= other.value
        return self

    def __rmul__(self, other: Number | Frame) -> Frame:
        if not isinstance(other, Frame):
            other = Frame(other)
        return Frame(other.value * self.value)

    def __floordiv__(self, other: Number | Frame) -> Frame:
        if not isinstance(other, Frame):
            other = Frame(other)
        return Frame(int(self.value // other.value))

    def __ifloordiv__(self, other: Number | Frame) -> Frame:
        if not isinstance(other, Frame):
            other = Frame(other)
        self.value = int(self.value // other.value)
        return self

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        try_load(
            state, 'value', int, self.__init__,  # type: ignore
            'Failed to load Frame instance'
        )


class Time(YAMLObjectWrapper):
    __slots__ = ('value', )

    def __init__(self, init_value: int | Time | timedelta | Frame | None = None, **kwargs: Any):
        if isinstance(init_value, int):
            init_value = Frame(init_value)

        if isinstance(init_value, timedelta):
            self.value = init_value
        elif isinstance(init_value, Time):
            self.value = init_value.value
        elif isinstance(init_value, Frame):
            self.value = main_window().current_output.to_time(init_value).value
        elif any(kwargs):
            self.value = timedelta(**kwargs)
        elif init_value is None:
            self.value = timedelta()
        else:
            raise TypeError

    def __add__(self, other: Time) -> Time:
        return Time(self.value + other.value)

    def __iadd__(self, other: Time) -> Time:
        self.value += other.value
        return self

    def __sub__(self, other: int | Time | timedelta | Frame | None) -> Time:
        if isinstance(other, Time):
            return Time(self.value - other.value)
        return self - Time(other)

    def __isub__(self, other: int | Time | timedelta | Frame | None) -> Time:
        if not isinstance(other, Time):
            other = Time(other)
        self.value -= other.value
        return self

    def __mul__(self, other: int) -> Time:
        return Time(self.value * other)

    def __imul__(self, other: int) -> Time:
        self.value *= other
        return self

    def __rmul__(self, other: int) -> Time:
        return Time(other * self.value)

    def __truediv__(self, other: float) -> Time:
        return Time(self.value / other)

    def __itruediv__(self, other: float) -> Time:
        self.value /= other
        return self

    def __str__(self) -> str:
        from ...utils import strfdelta

        return strfdelta(self, '%h:%M:%S.%Z')

    def __float__(self) -> float:
        return cast(float, self.value.total_seconds())

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        try_load(
            state, 'value', timedelta, self.__init__,  # type: ignore
            'Failed to load Time instance'
        )



from __future__ import annotations

from typing import Any

from PyQt6.QtCore import QTime, pyqtSignal
from PyQt6.QtWidgets import QTimeEdit, QWidget

from ..abstracts import SpinBox
from ..types import Frame, Time

__all__ = [
    'FrameEdit', 'TimeEdit'
]


def to_qtime(time: Time) -> QTime:
    seconds = time.value.seconds % (24 * 3600)
    hours = seconds // 3600
    seconds %= 3600
    minutes = seconds // 60
    seconds %= 60
    milliseconds = time.value.microseconds // 1000
    return QTime(hours, minutes, seconds, milliseconds)


def from_qtime(qtime: QTime, t: type[Time]) -> Time:
    return t(milliseconds=qtime.msecsSinceStartOfDay())


class FrameEdit(SpinBox):
    valueChanged = pyqtSignal(Frame, Frame)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.setMinimum(Frame(0))

        self.oldValue: Frame = self.value()
        super().valueChanged.connect(self._valueChanged)

    def _valueChanged(self, newValue: int) -> None:
        self.valueChanged.emit(self.value(), self.oldValue)

    def value(self) -> Frame:  # type: ignore[override]
        return Frame(super().value())

    def setValue(self, newValue: Frame) -> None:  # type: ignore[override]
        super().setValue(int(newValue))

    def minimum(self) -> Frame:  # type: ignore[override]
        return Frame(super().minimum())

    def setMinimum(self, newValue: Frame) -> None:  # type: ignore[override]
        super().setMinimum(int(newValue))

    def maximum(self) -> Frame:  # type: ignore[override]
        return Frame(super().maximum())

    def setMaximum(self, newValue: Frame) -> None:  # type: ignore[override]
        super().setMaximum(int(newValue))


class TimeEdit(QTimeEdit):
    valueChanged = pyqtSignal(Time, Time)

    def __init__(self, parent: QWidget | None = None, **kwargs: Any) -> None:
        super().__init__(parent, **kwargs, timeChanged=self._timeChanged)  # type: ignore

        self.setDisplayFormat('H:mm:ss.zzz')
        self.setButtonSymbols(QTimeEdit.ButtonSymbols.NoButtons)
        self.setMinimum(Time())

        self.oldValue: Time = self.value()

    def _timeChanged(self, newValue: QTime) -> None:
        self.valueChanged.emit(self.value(), self.oldValue)
        self.oldValue = self.value()

    def value(self) -> Time:
        return from_qtime(super().time(), Time)

    def setValue(self, newValue: Time) -> None:
        super().setTime(to_qtime(newValue))

    def minimum(self) -> Time:
        return from_qtime(super().minimumTime(), Time)

    def setMinimum(self, newValue: Time) -> None:
        super().setMinimumTime(to_qtime(newValue))

    def maximum(self) -> Time:
        return from_qtime(super().maximumTime(), Time)

    def setMaximum(self, newValue: Time) -> None:
        super().setMaximumTime(to_qtime(newValue))



from __future__ import annotations

from array import array
from math import floor
from typing import Any, Mapping, cast

import vapoursynth as vs
from PyQt6.QtCore import QIODevice
from PyQt6.QtMultimedia import QAudioFormat, QAudioOutput, QAudioSink

from ..abstracts import AbstractYAMLObject, main_window, try_load
from .units import Frame, Time

__all__ = [
    'AudioOutput'
]


class AudioOutput(AbstractYAMLObject):
    SAMPLES_PER_FRAME = 3072  # https://github.com/vapoursynth/vapoursynth/blob/master/include/VapourSynth4.h#L32

    storable_attrs = ('name', )

    __slots__ = (
        *storable_attrs, 'vs_output', 'index', 'fps_num', 'fps_den', 'format',
        'total_frames', 'total_time', 'end_frame', 'fps', 'is_mono',
        'source_vs_output', 'main', 'qformat', 'qoutput', 'iodevice', 'flags'
    )

    def __init__(self, vs_output: vs.AudioNode, index: int, new_storage: bool = False) -> None:
        self.setValue(vs_output, index, new_storage)

    def setValue(self, vs_output: vs.AudioNode, index: int, new_storage: bool = False) -> None:
        self.main = main_window()

        with self.main.env:
            vs_outputs = list(vs.get_outputs().values())

        self.index = vs_outputs.index(vs_output)
        self.source_vs_output = vs_output
        self.vs_output = self.source_vs_output
        self.is_mono = self.vs_output.num_channels == 1

        self.info = self.main.user_output_info[vs.AudioNode].get(self.index, {})

        (self.arrayType, sampleTypeQ) = (
            'f', QAudioFormat.SampleFormat.Float
        ) if self.vs_output.sample_type == vs.FLOAT else (
            'I' if self.vs_output.bits_per_sample <= 16 else 'L', QAudioFormat.SampleFormat.Int16
        )

        sample_size = 8 * self.vs_output.bytes_per_sample

        self.qformat = QAudioFormat()
        self.qformat.setChannelCount(self.vs_output.num_channels)
        self.qformat.setSampleRate(self.vs_output.sample_rate)
        self.qformat.setSampleFormat(sampleTypeQ)
        # self.qformat.setSampleSize(sample_size)
        # self.qformat.setByteOrder(Qt.LittleEndian)
        # self.qformat.setCodec('audio/pcm')

        self.qoutput = QAudioOutput(self.main)

        if not self.qoutput.device().isFormatSupported(self.qformat):
            raise RuntimeError('Audio format not supported')

        self.qaudiosink = QAudioSink(self.qoutput.device(), self.qformat, self.main)
        self.qaudiosink.setBufferSize(sample_size * self.SAMPLES_PER_FRAME)

        iodevice = cast(QIODevice | None, self.qaudiosink.start())

        if iodevice is None:
            from vstools import CustomRuntimeError

            raise CustomRuntimeError(
                'The current QT version has a bug for dll loading, you need to go into '
                'C:\\System32 and copy "mfplat.dll" into "mfplat.dll.dll".'
            )

        self.iodevice = iodevice

        self.fps_num = self.vs_output.sample_rate
        self.fps_den = self.SAMPLES_PER_FRAME
        self.fps = self.fps_num / self.fps_den
        self.total_frames = Frame(self.vs_output.num_frames)
        self.total_time = self.to_time(self.total_frames - Frame(1))

        self.audio_buffer = array(self.arrayType, [0] * self.SAMPLES_PER_FRAME * (self.vs_output.bytes_per_sample // 2))

        if not hasattr(self, 'name'):
            from ...models.outputs import AudioOutputs

            if vs_output in (vs_outputs := list(vs.get_outputs().values())):
                self.name = self.info.get('name', 'Track ' + str(self.index))
                if isinstance(self.main.toolbars.playback.audio_outputs, AudioOutputs):
                    self.main.toolbars.playback.audio_outputs.setData(
                        self.main.toolbars.playback.audio_outputs.index(index), self.name
                    )

    def clear(self) -> None:
        self.source_vs_output = self.vs_output = None  # type: ignore

    def render_audio_frame(self, frame: Frame) -> None:
        self.render_raw_audio_frame(self.vs_output.get_frame(int(frame)))

    def render_raw_audio_frame(self, vs_frame: vs.AudioFrame) -> None:
        if self.is_mono:
            self.iodevice.write(vs_frame[0].tobytes())
        else:
            self.audio_buffer[0::2] = array(self.arrayType, vs_frame[0].tobytes())
            self.audio_buffer[1::2] = array(self.arrayType, vs_frame[1].tobytes())

            self.iodevice.write(self.audio_buffer.tobytes())

    @property
    def volume(self) -> float:
        return self.qoutput.volume()

    @volume.setter
    def volume(self, newVolume: float) -> None:
        return self.qoutput.setVolume(newVolume)

    def _calculate_frame(self, seconds: float) -> int:
        return floor(seconds * self.fps)

    def _calculate_seconds(self, frame_num: int) -> float:
        return frame_num / self.fps

    def to_frame(self, time: Time) -> Frame:
        return Frame(self._calculate_frame(float(time)))

    def to_time(self, frame: Frame) -> Time:
        return Time(seconds=self._calculate_seconds(int(frame)))

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        try_load(state, 'name', str, self.__setattr__)



from __future__ import annotations

import logging
import re
from copy import deepcopy
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Mapping, cast

from PyQt6.QtCore import QKeyCombination, QModelIndex, Qt
from PyQt6.QtGui import QColor
from PyQt6.QtWidgets import QFileDialog, QLabel

from ...core import (
    AbstractToolbar, CheckBox, ComboBox, Frame, HBoxLayout, LineEdit, Notches, PushButton, Time, try_load
)
from ...models import SceningList, SceningLists
from ...utils import fire_and_forget, set_status_label
from .dialog import SceningListDialog
from .settings import SceningSettings

if TYPE_CHECKING:
    from ...main import MainWindow


__all__ = [
    'SceningToolbar'
]


class SceningToolbar(AbstractToolbar):
    storable_attrs = ('current_list_index', 'lists', 'first_frame', 'second_frame')

    __slots__ = (
        *storable_attrs[1:],
        'export_template_pattern', 'export_template_scenes_pattern',
        'scening_list_dialog', 'supported_file_types',
        'add_list_button', 'remove_list_button', 'view_list_button',
        'toggle_first_frame_button', 'toggle_second_frame_button',
        'add_single_frame_button',
        'add_to_list_button', 'remove_last_from_list_button',
        'export_multiline_button', 'export_template_lineedit',
        'always_show_scene_marks_checkbox',
        'status_label', 'import_file_button', 'items_combobox',
        'remove_at_current_frame_button',
        'seek_to_next_button', 'seek_to_prev_button'
    )

    settings: SceningSettings

    def __init__(self, main: MainWindow) -> None:
        super().__init__(main, SceningSettings(self))
        self.setup_ui()

        self.lists = SceningLists()

        self.first_frame: Frame | None = None
        self.second_frame: Frame | None = None
        self.export_template_pattern = re.compile(r'.*(?:{start}|{end}|{label}).*')
        self.export_template_scenes_pattern = re.compile(r'.+')

        self.items_combobox.setModel(self.lists)
        self.scening_update_status_label()
        self.scening_list_dialog = SceningListDialog(self.main)

        self.supported_file_types = {
            'Aegisub Project (*.ass)': self.import_ass,
            'AvsP Session (*.ses)': self.import_ses,
            'CUE Sheet (*.cue)': self.import_cue,
            'DGIndex Project (*.dgi)': self.import_dgi,
            'IfoEdit Celltimes (*.txt)': self.import_celltimes,
            'L-SMASH Works Index (*.lwi)': self.import_lwi,
            'Matroska Timestamps v1 (*.txt)': self.import_matroska_timestamps_v1,
            'Matroska Timestamps v2 (*.txt)': self.import_matroska_timestamps_v2,
            'Matroska Timestamps v3 (*.txt)': self.import_matroska_timestamps_v3,
            'Matroska XML Chapters (*.xml)': self.import_matroska_xml_chapters,
            'OGM Chapters (*.txt)': self.import_ogm_chapters,
            'TFM Log (*.txt)': self.import_tfm,
            'VSEdit Bookmarks (*.bookmarks)': self.import_vsedit,
            'x264/x265 2 Pass Log (*.log)': self.import_x264_2pass_log,
            'x264/x265 QP File (*.qp *.txt)': self.import_qp,
            'XviD Log (*.txt)': self.import_xvid,
            'Generic Mappings (*.txt)': self.import_generic,
        }

        self.items_combobox.valueChanged.connect(self.on_current_list_changed)
        self.add_list_button.clicked.connect(self.on_add_list_clicked)
        self.remove_list_button.clicked.connect(self.on_remove_list_clicked)
        self.view_list_button.clicked.connect(self.on_view_list_clicked)
        self.import_file_button.clicked.connect(self.on_import_file_clicked)
        self.seek_to_prev_button.clicked.connect(self.on_seek_to_prev_clicked)
        self.seek_to_next_button.clicked.connect(self.on_seek_to_next_clicked)

        self.add_single_frame_button.clicked.connect(self.on_add_single_frame_clicked)
        self.toggle_first_frame_button.clicked.connect(self.on_first_frame_clicked)
        self.toggle_second_frame_button.clicked.connect(self.on_second_frame_clicked)
        self.add_to_list_button.clicked.connect(self.on_add_to_list_clicked)
        self.remove_last_from_list_button.clicked.connect(self.on_remove_last_from_list_clicked)
        self.remove_at_current_frame_button.clicked.connect(self.on_remove_at_current_frame_clicked)
        self.export_template_lineedit.textChanged.connect(self.check_remove_export_possibility)
        self.export_multiline_button.clicked.connect(self.export_multiline)

        self.add_shortcuts()

        # FIXME: get rid of workaround
        self._on_list_items_changed = lambda *arg: self.on_list_items_changed(*arg)

        self.set_qobject_names()

    def setup_ui(self) -> None:
        super().setup_ui()

        self.items_combobox = ComboBox[SceningList](duplicatesEnabled=True, minimumContentsLength=4)

        self.add_list_button = PushButton('Add List')

        self.remove_list_button = PushButton('Remove List', enabled=False)

        self.view_list_button = PushButton('View List', enabled=False)

        self.import_file_button = PushButton('Import List')

        self.seek_to_prev_button = PushButton('⏪', enabled=False)

        self.seek_to_next_button = PushButton('⏩', enabled=False)

        self.always_show_scene_marks_checkbox = CheckBox('Always show scene marks in the timeline', checked=False)

        self.add_single_frame_button = PushButton('🆎', tooltip='Add Single Frame Scene')

        self.toggle_first_frame_button = PushButton('🅰️', tooltip='Toggle Start of New Scene', checkable=True)

        self.toggle_second_frame_button = PushButton('🅱️', tooltip='Toggle End of New Scene', checkable=True)

        self.label_lineedit = LineEdit('New Scene Label')

        self.add_to_list_button = PushButton('Add to List', enabled=False)

        self.remove_last_from_list_button = PushButton('Remove Last', enabled=False)

        self.remove_at_current_frame_button = PushButton('Remove at Current Frame', enabled=False)

        self.export_template_lineedit = LineEdit(
            'Export Template',
            text=self.settings.default_export_template,
            tooltip=(
                r'Use {start} and {end} as placeholders.'
                r'Both are valid for single frame scenes. '
                r'{label} is available, too. '
            )
        )

        self.export_multiline_button = PushButton('Export Multiline', enabled=False)

        HBoxLayout(self.vlayout, [
            self.items_combobox,
            self.add_list_button,
            self.remove_list_button,
            self.view_list_button,
            self.import_file_button,
            self.get_separator(),
            self.seek_to_prev_button,
            self.seek_to_next_button,
            self.always_show_scene_marks_checkbox
        ]).addStretch()

        HBoxLayout(self.vlayout, [
            self.add_single_frame_button,
            self.toggle_first_frame_button,
            self.toggle_second_frame_button,
            self.label_lineedit,
            self.add_to_list_button,
            self.remove_last_from_list_button,
            self.remove_at_current_frame_button,
            self.get_separator(),
            self.export_template_lineedit,
            self.export_multiline_button
        ]).addStretch(2)

        # statusbar label
        self.status_label = QLabel(self)
        self.status_label.setVisible(False)
        self.main.statusbar.addPermanentWidget(self.status_label)

    def add_shortcuts(self) -> None:
        for i, key in enumerate(self.num_keys[:-2]):
            self.add_shortcut(QKeyCombination(Qt.Modifier.SHIFT, key), partial(self.switch_list, i))  # type: ignore

        self.add_shortcut(
            QKeyCombination(Qt.Modifier.CTRL, Qt.Key.Key_Space).toCombined(), self.on_toggle_single_frame
        )
        self.add_shortcut(
            QKeyCombination(Qt.Modifier.CTRL, Qt.Key.Key_Left).toCombined(), self.seek_to_prev_button.click
        )
        self.add_shortcut(
            QKeyCombination(Qt.Modifier.CTRL, Qt.Key.Key_Right).toCombined(), self.seek_to_next_button.click
        )
        if self.main.settings.azerty_keybinds:
            self.add_shortcut(Qt.Key.Key_A, self.toggle_first_frame_button.click)
            self.add_shortcut(Qt.Key.Key_Z, self.toggle_second_frame_button.click)
        else:
            self.add_shortcut(Qt.Key.Key_Q, self.toggle_first_frame_button.click)
            self.add_shortcut(Qt.Key.Key_W, self.toggle_second_frame_button.click)
        self.add_shortcut(Qt.Key.Key_E, self.add_to_list_button.click)
        self.add_shortcut(Qt.Key.Key_R, self.remove_last_from_list_button.click)
        self.add_shortcut(
            QKeyCombination(Qt.Modifier.SHIFT, Qt.Key.Key_R).toCombined(), self.remove_at_current_frame_button.click
        )
        self.add_shortcut(
            Qt.Key.Key_B, lambda: self.scening_list_dialog.label_lineedit.setText(
                str(self.main.current_output.last_showed_frame)
            )
        )

    def on_toggle(self, new_state: bool) -> None:
        if new_state is True:
            self.check_add_to_list_possibility()
            self.check_remove_export_possibility()

        self.status_label.setVisible(self.is_notches_visible)

        super().on_toggle(new_state)

    def on_current_output_changed(self, index: int, prev_index: int) -> None:
        self.scening_list_dialog.on_current_output_changed(index, prev_index)

    def on_current_frame_changed(self, frame: Frame) -> None:
        self.check_remove_export_possibility()
        self.scening_list_dialog.on_current_frame_changed(frame, Time(frame))

    def get_notches(self) -> Notches:
        marks = Notches()
        if self.current_list is None:
            return marks
        for scene in self.current_list:
            marks.add(scene, cast(QColor, Qt.GlobalColor.green))

        return marks

    @property
    def current_list(self) -> SceningList | None:
        return self.items_combobox.currentValue()

    @property
    def current_list_index(self) -> int:
        return self.items_combobox.currentIndex()

    @current_list_index.setter
    def current_list_index(self, index: int) -> None:
        if (0 <= index < len(self.lists)):
            return self.items_combobox.setCurrentIndex(index)
        raise IndexError

    @property
    def is_notches_visible(self) -> bool:
        return self.always_show_scene_marks_checkbox.isChecked() or self.toggle_button.isChecked()

    # list management
    def on_add_list_clicked(self, checked: bool | None = None) -> None:
        _, self.current_list_index = self.lists.add()

    def on_current_list_changed(self, new_value: SceningList | None, old_value: SceningList) -> None:
        if new_value is not None:
            self.remove_list_button.setEnabled(True)
            self.view_list_button.setEnabled(True)
            new_value.rowsInserted.connect(self._on_list_items_changed)
            new_value.rowsRemoved.connect(self._on_list_items_changed)
            new_value.dataChanged.connect(self._on_list_items_changed)
            self.scening_list_dialog.on_current_list_changed(new_value)
        else:
            self.remove_list_button.setEnabled(False)
            self.view_list_button.setEnabled(False)

        if old_value is not None:
            try:
                old_value.rowsInserted.disconnect(self._on_list_items_changed)
                old_value.rowsRemoved.disconnect(self._on_list_items_changed)
                old_value.dataChanged.disconnect(self._on_list_items_changed)
            except (IndexError, TypeError):
                pass

        self.check_add_to_list_possibility()
        self.check_remove_export_possibility()
        self.notches_changed.emit(self)

    def on_list_items_changed(self, parent: QModelIndex, first: int, last: int) -> None:
        self.notches_changed.emit(self)

    def on_remove_list_clicked(self, checked: bool | None = None) -> None:
        self.lists.remove(self.current_list_index)

        if len(self.lists) == 0:
            self.remove_list_button.setEnabled(False)
            self.view_list_button.setEnabled(False)

    def on_view_list_clicked(self, checked: bool | None = None) -> None:
        self.scening_list_dialog.show()

    def switch_list(self, index: int) -> None:
        try:
            self.current_list_index = index
        except IndexError:
            pass

    # seeking
    def on_seek_to_prev_clicked(self, checked: bool | None = None) -> None:
        if self.current_list is None:
            return

        new_pos = self.current_list.get_prev_frame(self.main.current_output.last_showed_frame)
        if new_pos is None:
            return
        self.main.switch_frame(new_pos)

    def on_seek_to_next_clicked(self, checked: bool | None = None) -> None:
        if self.current_list is None:
            return

        new_pos = self.current_list.get_next_frame(self.main.current_output.last_showed_frame)
        if new_pos is None:
            return
        self.main.switch_frame(new_pos)

    # scene management
    def on_add_single_frame_clicked(self, checked: bool | None = None) -> None:
        if self.current_list is None:
            self.on_add_list_clicked()

        assert self.current_list is not None

        self.current_list.add(self.main.current_output.last_showed_frame, label=self.label_lineedit.text())

        self.check_remove_export_possibility()

    def on_add_to_list_clicked(self, checked: bool | None = None) -> None:
        self.current_list.add(self.first_frame, self.second_frame, self.label_lineedit.text())  # type: ignore

        if self.toggle_first_frame_button.isChecked():
            self.toggle_first_frame_button.click()
        if self.toggle_second_frame_button.isChecked():
            self.toggle_second_frame_button.click()
        self.add_to_list_button.setEnabled(False)
        self.label_lineedit.setText('')

        self.check_remove_export_possibility()

    def on_first_frame_clicked(self, checked: bool, frame: Frame | None = None) -> None:
        if frame is None:
            frame = self.main.current_output.last_showed_frame

        if checked:
            self.first_frame = frame
        else:
            self.first_frame = None
        self.scening_update_status_label()
        self.check_add_to_list_possibility()

    def on_remove_at_current_frame_clicked(self, checked: bool | None = None) -> None:
        if self.current_list is None:
            return

        curr = self.main.current_output.last_showed_frame

        for scene in self.current_list:
            if (scene.start == curr or scene.end == curr):
                self.current_list.remove(scene)

        self.remove_at_current_frame_button.clearFocus()
        self.check_remove_export_possibility()

    def on_remove_last_from_list_clicked(self, checked: bool | None = None) -> None:
        if self.current_list is None:
            return

        self.current_list.remove(self.current_list[-1])
        self.remove_last_from_list_button.clearFocus()
        self.check_remove_export_possibility()

    def on_second_frame_clicked(self, checked: bool, frame: Frame | None = None) -> None:
        if frame is None:
            frame = self.main.current_output.last_showed_frame

        if checked:
            self.second_frame = frame
        else:
            self.second_frame = None
        self.scening_update_status_label()
        self.check_add_to_list_possibility()

    def on_toggle_single_frame(self) -> None:
        if self.add_single_frame_button.isEnabled():
            self.add_single_frame_button.click()
        elif self.remove_at_current_frame_button.isEnabled():
            self.remove_at_current_frame_button.click()

    # import
    def on_import_file_clicked(self, checked: bool | None = None) -> None:
        filter_str = ';;'.join(self.supported_file_types.keys())
        path_strs, file_type = QFileDialog.getOpenFileNames(
            self.main, caption='Open chapters file', filter=filter_str
        )

        paths = [Path(path_str) for path_str in path_strs]
        for path in paths:
            self.import_file(self.supported_file_types[file_type], path)

    @fire_and_forget
    @set_status_label('Importing scening list')
    def import_file(self, import_func: Callable[[Path, SceningList, int], None], path: Path) -> None:
        out_of_range_count = 0
        scening_list, scening_list_index = self.lists.add(path.stem)

        import_func(path, scening_list, out_of_range_count)

        if out_of_range_count > 0:
            logging.warning(
                f'Scening import: {out_of_range_count} scenes were out of range of output, so they were dropped.')
        if len(scening_list) == 0:
            logging.warning(f"Scening import: nothing was imported from '{path.name}'.")
            self.lists.remove(scening_list_index)
        else:
            self.current_list_index = scening_list_index

    def import_ass(self, path: Path, scening_list: SceningList, out_of_range_count: int) -> None:
        '''
        Imports lines as scenes.
        Text is ignored.
        '''
        try:
            from pysubs2 import load as pysubs2_load  # type: ignore[import]
        except ModuleNotFoundError:
            raise RuntimeError(
                'vspreview: Can\'t import scenes from ass file, you\'re missing the `pysubs2` package!'
            )

        subs = pysubs2_load(str(path))
        for line in subs:
            t_start = Time(milliseconds=line.start)
            t_end = Time(milliseconds=line.end)
            try:
                scening_list.add(Frame(t_start), Frame(t_end))
            except ValueError:
                out_of_range_count += 1

    def import_celltimes(self, path: Path, scening_list: SceningList, out_of_range_count: int) -> None:
        '''
        Imports cell times as single-frame scenes
        '''

        for line in path.read_text('utf8').splitlines():
            try:
                scening_list.add(Frame(int(line)))
            except ValueError:
                out_of_range_count += 1

    def import_cue(self, path: Path, scening_list: SceningList, out_of_range_count: int) -> None:
        '''
        Imports tracks as scenes.
        Uses TITLE for scene label.
        '''
        try:
            from cueparser import CueSheet  # type: ignore[import]
        except ModuleNotFoundError:
            raise RuntimeError(
                'vspreview: Can\'t import scenes from cue file, you\'re missing the `cueparser` package!'
            )

        def offset_to_time(offset: str) -> Time | None:
            pattern = re.compile(r'(\d{1,2}):(\d{1,2}):(\d{1,2})')
            match = pattern.match(offset)
            if match is None:
                return None
            return Time(minutes=int(match[1]), seconds=int(match[2]), milliseconds=int(match[3]) / 75 * 1000)

        cue_sheet = CueSheet()
        cue_sheet.setOutputFormat('')
        cue_sheet.setData(path.read_text('utf8'))
        cue_sheet.parse()

        for track in cue_sheet.tracks:
            if track.offset is None:
                continue
            offset = offset_to_time(track.offset)
            if offset is None:
                logging.warning(f"Scening import: INDEX timestamp '{track.offset}' format isn't supported.")
                continue
            start = Frame(offset)

            end = None
            if track.duration is not None:
                end = Frame(offset + Time(track.duration))

            label = ''
            if track.title is not None:
                label = track.title

            try:
                scening_list.add(start, end, label)
            except ValueError:
                out_of_range_count += 1

    def import_dgi(self, path: Path, scening_list: SceningList, out_of_range_count: int) -> None:
        '''
        Imports IDR frames as single-frame scenes.
        '''
        pattern = re.compile(r'IDR\s\d+\n(\d+):FRM', re.RegexFlag.MULTILINE)
        for match in pattern.findall(path.read_text('utf8')):
            try:
                scening_list.add(Frame(match))
            except ValueError:
                out_of_range_count += 1

    def import_lwi(self, path: Path, scening_list: SceningList, out_of_range_count: int) -> None:
        '''
        Imports Key=1 frames as single-frame scenes.
        Ignores everything besides Index=0 video stream.
        '''
        AV_CODEC_ID_FIRST_AUDIO = 0x10000
        STREAM_INDEX = 0
        IS_KEY = 1

        pattern = re.compile(r'Index={}.*?Codec=(\d+).*?\n.*?Key=(\d)'.format(
            STREAM_INDEX
        ))

        frame = Frame(0)
        for match in pattern.finditer(path.read_text('utf8'), re.RegexFlag.MULTILINE):
            if int(match[1]) >= AV_CODEC_ID_FIRST_AUDIO:
                frame += Frame(1)
                continue

            if not int(match[2]) == IS_KEY:
                frame += Frame(1)
                continue

            try:
                scening_list.add(deepcopy(frame))
            except ValueError:
                out_of_range_count += 1

            frame += Frame(1)

    def import_matroska_xml_chapters(self, path: Path, scening_list: SceningList, out_of_range_count: int) -> None:
        '''
        Imports chapters as scenes.
        Preserve end time and text if they're present.
        '''
        from xml.etree import ElementTree

        timestamp_pattern = re.compile(r'(\d{2}):(\d{2}):(\d{2}(?:\.\d{3})?)')

        try:
            root = ElementTree.parse(str(path)).getroot()
        except ElementTree.ParseError as exc:
            logging.warning(f"Scening import: error occurred while parsing '{path.name}':")
            logging.warning(exc.msg)
            return
        for chapter in root.iter('ChapterAtom'):
            start_element = chapter.find('ChapterTimeStart')
            if start_element is None or start_element.text is None:
                continue
            match = timestamp_pattern.match(start_element.text)
            if match is None:
                continue
            start = Frame(Time(hours=int(match[1]), minutes=int(match[2]), seconds=float(match[3])))

            end = None
            end_element = chapter.find('ChapterTimeEnd')
            if end_element is not None and end_element.text is not None:
                match = timestamp_pattern.match(end_element.text)
                if match is not None:
                    end = Frame(Time(hours=int(match[1]), minutes=int(match[2]), seconds=float(match[3])))

            label = ''
            label_element = chapter.find('ChapterDisplay/ChapterString')
            if label_element is not None and label_element.text is not None:
                label = label_element.text

            try:
                scening_list.add(start, end, label)
            except ValueError:
                out_of_range_count += 1

    def import_ogm_chapters(self, path: Path, scening_list: SceningList, out_of_range_count: int) -> None:
        '''
        Imports chapters as signle-frame scenes.
        Uses NAME for scene label.
        '''
        pattern = re.compile(
            r'(CHAPTER\d+)=(\d{2}):(\d{2}):(\d{2}(?:\.\d{3})?)\n\1NAME=(.*)',
            re.RegexFlag.MULTILINE
        )
        for match in pattern.finditer(path.read_text('utf8')):
            time = Time(hours=int(match[2]), minutes=int(match[3]), seconds=float(match[4]))
            try:
                scening_list.add(Frame(time), label=match[5])
            except ValueError:
                out_of_range_count += 1

    def import_qp(self, path: Path, scening_list: SceningList, out_of_range_count: int) -> None:
        '''
        Imports I- and K-frames as single-frame scenes.
        '''
        pattern = re.compile(r'(\d+)\sI|K')
        for match in pattern.findall(path.read_text('utf8')):
            try:
                scening_list.add(Frame(int(match)))
            except ValueError:
                out_of_range_count += 1

    def import_ses(self, path: Path, scening_list: SceningList, out_of_range_count: int) -> None:
        '''
        Imports bookmarks as single-frame scenes
        '''
        import pickle

        with path.open('rb') as f:
            try:
                session = pickle.load(f)
            except pickle.UnpicklingError:
                logging.warning('Scening import: failed to load .ses file.')
                return
        if 'bookmarks' not in session:
            return

        for bookmark in session['bookmarks']:
            try:
                scening_list.add(Frame(bookmark[0]))
            except ValueError:
                out_of_range_count += 1

    def import_matroska_timestamps_v1(self, path: Path, scening_list: SceningList, out_of_range_count: int) -> None:
        '''
        Imports listed scenes.
        Uses FPS for scene label.
        '''
        pattern = re.compile(r'(\d+),(\d+),(\d+(?:\.\d+)?)')

        for match in pattern.finditer(path.read_text('utf8')):
            try:
                scening_list.add(
                    Frame(int(match[1])), Frame(int(match[2])), '{:.3f} fps'.format(float(match[3]))
                )
            except ValueError:
                out_of_range_count += 1

    def import_matroska_timestamps_v2(self, path: Path, scening_list: SceningList, out_of_range_count: int) -> None:
        '''
        Imports intervals of constant FPS as scenes.
        Uses FPS for scene label.
        '''
        timestamps = list[Time]()
        for line in path.read_text('utf8').splitlines():
            try:
                timestamps.append(Time(milliseconds=float(line)))
            except ValueError:
                continue

        if len(timestamps) < 2:
            logging.warning(
                "Scening import: timestamps file contains less than 2 timestamps, so there's nothing to import."
            )
            return

        deltas = [
            timestamps[i] - timestamps[i - 1]
            for i in range(1, len(timestamps))
        ]
        scene_delta = deltas[0]
        scene_start = Frame(0)
        scene_end: Frame | None = None
        for i in range(1, len(deltas)):
            if abs(round(float(deltas[i] - scene_delta), 6)) <= 0.000_001:
                continue
            # TODO: investigate, why offset by -1 is necessary here
            scene_end = Frame(i - 1)
            try:
                scening_list.add(scene_start, scene_end, '{:.3f} fps'.format(1 / float(scene_delta)))
            except ValueError:
                out_of_range_count += 1
            scene_start = Frame(i)
            scene_end = None
            scene_delta = deltas[i]

        if scene_end is None:
            try:
                scening_list.add(
                    scene_start, Frame(len(timestamps) - 1),
                    '{:.3f} fps'.format(1 / float(scene_delta))
                )
            except ValueError:
                out_of_range_count += 1

    def import_matroska_timestamps_v3(self, path: Path, scening_list: SceningList, out_of_range_count: int) -> None:
        '''
        Imports listed scenes, ignoring gaps.
        Uses FPS for scene label.
        '''
        pattern = re.compile(
            r'^((?:\d+(?:\.\d+)?)|gap)(?:,\s?(\d+(?:\.\d+)?))?',
            re.RegexFlag.MULTILINE
        )

        assume_pattern = re.compile(r'assume (\d+(?:\.\d+))')
        if len(mmatch := assume_pattern.findall(path.read_text('utf8'))) > 0:
            default_fps = float(mmatch[0])
        else:
            logging.warning('Scening import: "assume" entry not found.')
            return

        pos = Time()
        for match in pattern.finditer(path.read_text('utf8')):
            if match[1] == 'gap':
                pos += Time(seconds=float(match[2]))
                continue

            interval = Time(seconds=float(match[1]))
            fps = float(match[2]) if (match.lastindex or 0) >= 2 else default_fps

            try:
                scening_list.add(Frame(pos), Frame(pos + interval), '{:.3f} fps'.format(fps))
            except ValueError:
                out_of_range_count += 1

            pos += interval

    def import_tfm(self, path: Path, scening_list: SceningList, out_of_range_count: int) -> None:
        '''
        Imports TFM's 'OVR HELP INFORMATION'.
        Single combed frames are put into single-frame scenes.
        Frame groups are put into regular scenes.
        Combed probability is used for label.
        '''
        class TFMFrame(Frame):
            mic: int | None

        tfm_frame_pattern = re.compile(r'(\d+)\s\((\d+)\)')
        tfm_group_pattern = re.compile(r'(\d+),(\d+)\s\((\d+(?:\.\d+)%)\)')

        log = path.read_text('utf8')

        start_pos = log.find('OVR HELP INFORMATION')
        if start_pos == -1:
            logging.warning("Scening import: TFM log doesn't contain OVR Help Information.")
            return
        log = log[start_pos:]

        tfm_frames = set[TFMFrame]()
        for match in tfm_frame_pattern.finditer(log):
            tfm_frame = TFMFrame(int(match[1]))
            tfm_frame.mic = int(match[2])
            tfm_frames.add(tfm_frame)

        for match in tfm_group_pattern.finditer(log):
            try:
                scene = scening_list.add(Frame(int(match[1])), Frame(int(match[2])), f'{match[3]} combed')
            except ValueError:
                out_of_range_count += 1
                continue

            tfm_frames -= set(range(int(scene.start), int(scene.end) + 1))

        for tfm_frame in tfm_frames:
            try:
                scening_list.add(tfm_frame, label=str(tfm_frame.mic))
            except ValueError:
                out_of_range_count += 1

    def import_vsedit(self, path: Path, scening_list: SceningList, out_of_range_count: int) -> None:
        '''
        Imports bookmarks as single-frame scenes
        '''

        frames = []

        for bookmark in path.read_text('utf8').split(', '):
            try:
                frames.append(int(bookmark))
            except ValueError:
                out_of_range_count += 1

        ranges = list[list[int]]()
        prev_x: int
        for x in frames:
            if not ranges:
                ranges.append([x])
            elif x - prev_x == 1:
                ranges[-1].append(x)
            else:
                ranges.append([x])
            prev_x = int(x)

        for rang in ranges:
            scening_list.add(
                Frame(rang[0]),
                Frame(rang[-1]) if len(rang) > 1 else None
            )

    def import_x264_2pass_log(self, path: Path, scening_list: SceningList, out_of_range_count: int) -> None:
        '''
        Imports I- and K-frames as single-frame scenes.
        '''
        pattern = re.compile(r'in:(\d+).*type:I|K')
        for match in pattern.findall(path.read_text('utf8')):
            try:
                scening_list.add(Frame(int(match)))
            except ValueError:
                out_of_range_count += 1

    def import_xvid(self, path: Path, scening_list: SceningList, out_of_range_count: int) -> None:
        '''
        Imports I-frames as single-frame scenes.
        '''
        for i, line in enumerate(path.read_text('utf8').splitlines()):
            if not line.startswith('i'):
                continue
            try:
                scening_list.add(Frame(i - 3))
            except ValueError:
                out_of_range_count += 1

    def import_generic(self, path: Path, scening_list: SceningList, out_of_range_count: int) -> None:
        '''
        Import generic (rfs style) frame mappings: {start end}

        '''
        for line in path.read_text('utf8').splitlines():
            try:
                fnumbers = [int(n) for n in line.split()]
                scening_list.add(Frame(fnumbers[0]), Frame(fnumbers[1]))
            except ValueError:
                out_of_range_count += 1

    # export
    def export_multiline(self, checked: bool | None = None) -> None:
        if self.current_list is None:
            return

        template = self.export_template_lineedit.text()
        export_str = str()

        try:
            for scene in self.current_list:
                export_str += template.format(
                    start=scene.start, end=scene.end, label=scene.label, script_name=self.main.script_path.stem
                ) + '\n'
        except KeyError:
            logging.warning('Scening: export template contains invalid placeholders.')
            self.main.show_message('Export template contains invalid placeholders.')
            return

        self.main.clipboard.setText(export_str)
        self.main.show_message('Scening data exported to the clipboard')

    # misc
    def check_add_to_list_possibility(self) -> None:
        self.add_to_list_button.setEnabled(False)

        if not (self.current_list_index != -1 and (self.first_frame is not None or self.second_frame is not None)):
            return

        self.add_to_list_button.setEnabled(True)

    def check_remove_export_possibility(self, checked: bool | None = None) -> None:
        is_enabled = self.current_list is not None and len(self.current_list) > 0
        self.remove_last_from_list_button.setEnabled(is_enabled)
        self.seek_to_next_button.setEnabled(is_enabled)
        self.seek_to_prev_button.setEnabled(is_enabled)

        curr = self.main.current_output.last_showed_frame

        is_enabled = self.current_list is not None and curr in self.current_list
        self.add_single_frame_button.setEnabled(not is_enabled)
        self.remove_at_current_frame_button.setEnabled(is_enabled)

        is_enabled = self.export_template_pattern.fullmatch(self.export_template_lineedit.text()) is not None
        self.export_multiline_button.setEnabled(is_enabled)

    def scening_update_status_label(self) -> None:
        first_frame_text = str(self.first_frame) if self.first_frame is not None else ''
        second_frame_text = str(self.second_frame) if self.second_frame is not None else ''
        self.status_label.setText(f'Scening: {first_frame_text} - {second_frame_text} ')

    def __getstate__(self) -> Mapping[str, Any]:
        return super().__getstate__() | {
            'label': self.label_lineedit.text(),
            'scening_export_template': self.export_template_lineedit.text(),
            'always_show_scene_marks': self.always_show_scene_marks_checkbox.isChecked(),
        }

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        try_load(state, 'lists', SceningLists, self.__setattr__)
        try_load(state, 'current_list_index', int, self.current_list_index)
        try_load(state, 'first_frame', Frame, self.__setattr__, nullable=True)
        try_load(state, 'second_frame', Frame, self.__setattr__, nullable=True)

        if self.first_frame is not None:
            self.toggle_first_frame_button.setChecked(True)

        if self.second_frame is not None:
            self.toggle_second_frame_button.setChecked(True)

        self.scening_update_status_label()
        self.check_add_to_list_possibility()

        try_load(state, 'label', str, self.label_lineedit.setText)
        self.items_combobox.setModel(self.lists)

        try_load(state, 'scening_export_template', str, self.export_template_lineedit.setText)

        always_show_scene_marks = None
        try_load(state, 'always_show_scene_marks', bool, always_show_scene_marks, nullable=True)
        if always_show_scene_marks is None:
            always_show_scene_marks = self.settings.always_show_scene_marks

        self.always_show_scene_marks_checkbox.setChecked(always_show_scene_marks)
        self.status_label.setVisible(always_show_scene_marks)

        super().__setstate__(state)
